/***************************************************************************
                         mmx_imdft.c  -  description
                             -------------------
   MMX version of the AC-3 IMDFT process. Suitable for all processors 
   (PMMX, P2, and later) that support the MMX instruction set.
                             -------------------
    begin                : Tue Nov 29 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code     src         dest */
# define STAGE_OUTPUT_2R2I(inc, r0, op0, r1, op1, off) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r0 "), %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "movq " #off"(%" #r1 "), %%mm5\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "movq "     "%%mm5, "   "%%mm4\n\t" \
      "psrlq "    "$32, "     "%%mm0\n\t" \
      "pand "     "%%mm6, "   "%%mm1\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "psllq "    "$32, "     "%%mm4\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm3, "   "%%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm5\n\t" \
      "paddw "    "%%mm1, "   "%%mm3\n\t" \
      "psubw "    "%%mm2, "   "%%mm1\n\t" \
      "movq "     "(%" #r1 "), %%mm2\n\t" \
      "paddw "    "%%mm4, "   "%%mm3\n\t" \
      "psubw "    "%%mm4, "   "%%mm1\n\t" \
      "movq "     "%%mm3, "   "(%" #r0 ")\n\t" \
      "movq "     "%%mm2, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "por "      "%%mm3, "   "%%mm5\n\t" \
      "movq "     "%%mm2, "   "%%mm4\n\t" \
      "paddw "    "%%mm5, "   "%%mm2\n\t" \
      "psubw "    "%%mm4, "   "%%mm5\n\t" \
      "psubw "    "%%mm0, "   "%%mm2\n\t" \
      "paddw "    "%%mm0, "   "%%mm5\n\t" \
      "movq "     "%%mm2, "   #off "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm5, "   "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code     src         dest */
# define STAGE_OUTPUT___2I(inc, r0, op0, r1, op1, off) \
      "movq "     "(%" #r0 "), %%mm2\n\t" \
      "movq " #off"(%" #r0 "), %%mm1\n\t" \
      "movq "     "%%mm2, "   "%%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm1\n\t" \
      "movq "     "(%" #r1 "), %%mm4\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "movq " #off"(%" #r1 "), %%mm5\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "movq "     "%%mm5, "   "%%mm2\n\t" \
      "psubw "    "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm4, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm4\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "por "      "%%mm2, "   "%%mm4\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "pand "     "%%mm6, "   "%%mm3\n\t" \
      "psubw "    "%%mm3, "   "%%mm4\n\t" \
      "paddw "    "%%mm3, "   "%%mm5\n\t" \
      "movq "     "%%mm4, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm5, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code     src         dest */
# define STAGE_OUTPUT_2R__(inc, r0, op0, r1, op1, off) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r0 "), %%mm3\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm3, "   "%%mm1\n\t" \
      "movq " #off"(%" #r1 "), %%mm5\n\t" \
      "psrlq "    "$32, "     "%%mm2\n\t" \
      "movq "     "%%mm5, "   "%%mm4\n\t" \
      "pand "     "%%mm7, "   "%%mm0\n\t" \
      "psllq "    "$32, "     "%%mm4\n\t" \
      "pand "     "%%mm6, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm5\n\t" \
      "por "      "%%mm3, "   "%%mm0\n\t" \
      "movq "     "(%" #r1 "), %%mm3\n\t" \
      "paddw "    "%%mm4, "   "%%mm0\n\t" \
      "psubw "    "%%mm4, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "movq "     "%%mm3, "   "%%mm4\n\t" \
      "psubw "    "%%mm2, "   "%%mm5\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "pand "     "%%mm7, "   "%%mm4\n\t" \
      "paddw "    "%%mm2, "   "%%mm3\n\t" \
      "paddw "    "%%mm4, "   "%%mm5\n\t" \
      "movq "     "%%mm3, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm5, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %4    = 0x5A815A81
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code     src         dest */
# define SR_BF_k_eq_N8_2R(inc, r0, op0, r1, op1, off) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r1 "), %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm5\n\t" \
      "movq " #off"(%" #r0 "), %%mm2\n\t" \
      "psllq "    "$32, "     "%%mm5\n\t" \
      "movq "     "(%" #r1 "), %%mm3\n\t" \
      "psrlq "    "$32, "     "%%mm4\n\t" \
      "paddw "    "%%mm5, "   "%%mm0\n\t" \
      "movq "     "%%mm2, "   "%%mm5\n\t" \
      "psubw "    "%%mm4, "   "%%mm1\n\t" \
      "psrlq "    "$32, "     "%%mm5\n\t" \
      "movq "     "%%mm3, "   "%%mm4\n\t" \
      "paddw "    "%%mm5, "   "%%mm3\n\t" \
      "psllq "    "$32, "     "%%mm4\n\t" \
      "movq "     "%%mm0, "   "%%mm5\n\t" \
      "psubw "    "%%mm4, "   "%%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm5\n\t" \
      "movq "     "%%mm2, "   "%%mm4\n\t" \
      "psubw "    "%%mm5, "   "%%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm5\n\t" \
      "paddw "    "%%mm4, "   "%%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm5\n\t" \
      "movq "     "%%mm3, "   "%%mm4\n\t" \
      "psubw "    "%%mm5, "   "%%mm3\n\t" \
      "pand "     "%%mm7, "   "%%mm4\n\t" \
      "movd "     "%4, "      "%%mm5\n\t" \
      "paddw "    "%%mm4, "   "%%mm1\n\t" \
      "movq "     "%%mm3, "   "%%mm4\n\t" \
      "pmulhw "   "%%mm5, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm4\n\t" \
      "psllw "    "$1, "      "%%mm3\n\t" \
      "por "      "%%mm4, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   "%%mm4\n\t" \
      "pmulhw "   "%%mm5, "   "%%mm1\n\t" \
      "pand "     "%%mm6, "   "%%mm4\n\t" \
      "psllw "    "$1, "      "%%mm1\n\t" \
      "por "      "%%mm4, "   "%%mm1\n\t" \
      "psllq "    "$32, "     "%%mm5\n\t" \
      "movq "     "%%mm0, "   "%%mm4\n\t" \
      "pmulhw "   "%%mm5, "   "%%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm4\n\t" \
      "psllw "    "$1, "      "%%mm0\n\t" \
      "por "      "%%mm4, "   "%%mm0\n\t" \
      "movq "     "%%mm2, "   "%%mm4\n\t" \
      "pmulhw "   "%%mm5, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm4\n\t" \
      "psrlw "    "$14, "     "%%mm5\n\t" \
      "psllw "    "$1, "      "%%mm2\n\t" \
      "por "      "%%mm5, "   "%%mm0\n\t" \
      "por "      "%%mm4, "   "%%mm2\n\t" \
      "movq "     "%%mm5, "   "%%mm4\n\t" \
      "por "      "%%mm5, "   "%%mm2\n\t" \
      "psrlq "    "$32, "     "%%mm4\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "por "      "%%mm4, "   "%%mm3\n\t" \
      "movq "     "%%mm2, "   #off "(%" #r0 ")\n\t" \
      "por "      "%%mm4, "   "%%mm1\n\t" \
      "movq "     "%%mm3, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %2    = IDFT_CONST
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code     src         dest */
# define SR_BF_2R(inc, r0, op0, r1, op1, off, h1, h2) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r1 "), %%mm1\n\t" \
      "psrlq "    "$32, "     "%%mm0\n\t" \
      "movq "     "(%2), "    "%%mm4\n\t" \
      "punpcklwd ""%%mm0, "   "%%mm1\n\t" \
      "movq "     "%%mm4, "   "%%mm5\n\t" \
"punpck" #h1 "dq ""%%mm4, "   "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm0\n\t" \
"punpck" #h2 "dq ""%%mm5, "   "%%mm5\n\t" \
      "movq " #off"(%" #r0 "), %%mm2\n\t" \
      "pmaddwd "  "%%mm4, "   "%%mm0\n\t" \
      "movq "     "(%" #r1 "), %%mm3\n\t" \
      "pmaddwd "  "%%mm5, "   "%%mm1\n\t" \
      "movq "     "8(%2), "   "%%mm5\n\t" \
      "psrlq "    "$32, "     "%%mm2\n\t" \
      "movq "     "%%mm5, "   "%%mm4\n\t" \
"punpck" #h2 "dq ""%%mm5, "   "%%mm5\n\t" \
      "punpcklwd ""%%mm2, "   "%%mm3\n\t" \
"punpck" #h1 "dq ""%%mm4, "   "%%mm4\n\t" \
      "movq "     "%%mm3, "   "%%mm2\n\t" \
      "pmaddwd "  "%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm7, "   "%%mm5\n\t" \
      "pmaddwd "  "%%mm4, "   "%%mm2\n\t" \
      "punpcklwd ""%%mm6, "   "%%mm5\n\t" \
      "psrlw "    "$2, "      "%%mm5\n\t" \
      "paddd "    "%%mm5, "   "%%mm2\n\t" \
      "paddd "    "%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm2, "   "%%mm4\n\t" \
      "psubd "    "%%mm0, "   "%%mm2\n\t" \
      "paddd "    "%%mm4, "   "%%mm0\n\t" \
      "psrad "    "$15, "     "%%mm2\n\t" \
      "movq "     "%%mm3, "   "%%mm5\n\t" \
      "psrad "    "$15, "     "%%mm0\n\t" \
      "psubd "    "%%mm1, "   "%%mm3\n\t" \
      "paddd "    "%%mm5, "   "%%mm1\n\t" \
      "psrad "    "$15, "     "%%mm3\n\t" \
      "pxor "     "%%mm4, "   "%%mm4\n\t" \
      "psrad "    "$15, "     "%%mm1\n\t" \
      "pxor "     "%%mm5, "   "%%mm5\n\t" \
      "packssdw " "%%mm4, "   "%%mm3\n\t" \
      "packssdw " "%%mm2, "   "%%mm5\n\t" \
      "movq "     "(%" #r0 "), %%mm2\n\t" \
      "packssdw " "%%mm4, "   "%%mm1\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "packssdw " "%%mm0, "   "%%mm4\n\t" \
      "movq " #off"(%" #r0 "), %%mm0\n\t" \
      "por "      "%%mm5, "   "%%mm2\n\t" \
      "movq "     "(%" #r1 "), %%mm5\n\t" \
      "pand "     "%%mm7, "   "%%mm0\n\t" \
      "movq "     "%%mm2, "   "(%" #r0 ")\n\t" \
      "pand "     "%%mm6, "   "%%mm5\n\t" \
      "movq " #off"(%" #r1 "), %%mm2\n\t" \
      "por "      "%%mm4, "   "%%mm0\n\t" \
      "por "      "%%mm1, "   "%%mm5\n\t" \
      "movq "     "%%mm0, "   #off "(%" #r0 ")\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "movq "     "%%mm5, "   "(%" #r1 ")\n\t" \
      "por "      "%%mm3, "   "%%mm2\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm2, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code     src         dest */
# define PRE_POST_SCALING_SETUP \
      "movq "     "%%mm7, "   "%%mm3\n\t" \
      "punpcklwd ""%%mm6, "   "%%mm3\n\t" \
      "psrlw "    "$2, "      "%%mm3\n\t"

   /* registers used: %0   = data+? -> add(4)
                      %1   = data+? -> sub(4)
                      %5   = MMX_256_CONST
                      %%mm3 = 0x00003FFF00003FFF
                      %%mm6 = 0xFFFFFFFF00000000
                      %%mm7 = 0x00000000FFFFFFFF */
   /*  op_code     src         dest */
# define PRE_POST_SCALING(op) \
      "movq "     "(%0), "    "%%mm0\n\t" \
      "movq "     "(%1), "    "%%mm1\n\t" \
      "psllq "    "$32, "     "%%mm0\n\t" \
      "movq "     "(%5), "    "%%mm4\n\t" \
      "punpckhwd ""%%mm1, "   "%%mm0\n\t" \
      "movq "     "%%mm4, "   "%%mm5\n\t" \
      "punpckldq ""%%mm4, "   "%%mm4\n\t" \
      "movq "     "%%mm0, "   "%%mm1\n\t" \
      "punpckhdq ""%%mm5, "   "%%mm5\n\t" \
      "movq "     "(%0), "    "%%mm2\n\t" \
      "pmaddwd "  "%%mm4, "   "%%mm0\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "pmaddwd "  "%%mm5, "   "%%mm1\n\t" \
      "paddd "    "%%mm3, "   "%%mm0\n\t" \
      "p" #op "d ""%%mm3, "   "%%mm1\n\t" \
      "psrad "    "$15, "     "%%mm0\n\t" \
      "movq "     "(%1), "    "%%mm5\n\t" \
      "pxor "     "%%mm4, "   "%%mm4\n\t" \
      "psrad "    "$15, "     "%%mm1\n\t" \
      "pand "     "%%mm7, "   "%%mm5\n\t" \
      "packssdw " "%%mm4, "   "%%mm0\n\t" \
      "addl "     "$8, "      "%5\n\t" \
      "packssdw " "%%mm1, "   "%%mm4\n\t" \
      "por "      "%%mm2, "   "%%mm0\n\t" \
      "p" #op "d ""%%mm4, "   "%%mm5\n\t" \
      "movq "     "%%mm0, "   "(%0)\n\t" \
      "addl "     "$8, "      "%0\n\t" \
      "movq "     "%%mm5, "   "(%1)\n\t" \
      "subl "     "$8, "      "%1\n\t"


void AC3_imdft256(AC3_COEFF *data)
{
   __asm__ __volatile__(
   /* INIT MASKS */
   /*  op_code     src         dest */
      "pcmpeqd "  "%%mm7, "   "%%mm7\n\t"
      "pxor "     "%%mm6, "   "%%mm6\n\t"
      "movq "     "%%mm7, "   "%%mm5\n\t"
      "punpckldq ""%%mm6, "   "%%mm7\n\t"    /* mm7 = 0x00000000FFFFFFFF */
      "punpckldq ""%%mm5, "   "%%mm6\n\t"    /* mm6 = 0xFFFFFFFF00000000 */

   /* Pre-IDFT Scaling:
      registers used: %0 = data
                      %1 = data+252
                      %3 = 64 */
   /*  op_code     src         dest */
      PRE_POST_SCALING_SETUP
      /* %0 = data      %1 = data+252 */
      "PRE_LOOP_256:\n\t"
      PRE_POST_SCALING(add)
      "dec "                  "%3\n\t"
      "jnz "      "PRE_LOOP_256\n\t"
      /* %0 = data+256  %1 = data-4 */

   /* STAGE 1: 2pt-DFTs
      registers used: %0 = data
                      %1 = data+128
                      %3 = 32 */
   /*  op_code     src         dest */
      "subl "     "$(256*2), ""%0\n\t"
      "addl "     "$(132*2), ""%1\n\t"
      "addl "     "$32, "     "%3\n\t"
      /* %0 = data      %1 = data+128 */
      "STAGE1_LOOP1_256:\n\t"
      STAGE1
      "dec "                  "%3\n\t"
      "jnz "      "STAGE1_LOOP1_256\n\t"
      /* %0 = data+128  %1 = data+256 */

   /* STAGE 2: 4pt-DFTs
      registers used: %0 = data
                      %1 = data+188
                      %3 = 4 */
   /*  op_code     src         dest */
      "subl "     "$(128*2), ""%0\n\t"
      "subl "     "$(68*2), " "%1\n\t"
      "addl "     "$4, "      "%3\n\t"
      /* %0 = data      %1 = data+188 */
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT___2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      "STAGE2_LOOP1_256:\n\t"
      STAGE_OUTPUT___2I(4, 0, add, 1, sub, 128)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE2_LOOP1_256\n\t"
      "addl "     "$4, "      "%3\n\t"
      "STAGE2_LOOP2_256:\n\t"
      STAGE_OUTPUT_2R__(4, 0, add, 1, sub, 128)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE2_LOOP2_256\n\t"
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R__(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R2I(4, 0, sub, 1, add, 128)
      /* %0 = data+56   %1 = data+132 */

   /* STAGE 3: 8pt-DFTs
      registers used: %0 = data+56
                      %1 = data+132
                      %3 = 4
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "addl "     "$4, "      "%3\n\t"
      /* 1, 3 */
      /* %0 = data+56   %1 = data+132 */
      SR_BF_k_eq_0_2I(12, 0, sub, 1, add, 128)
      "STAGE3_LOOP1_256:\n\t"
      SR_BF_k_eq_0_2I( 4, 0, sub, 1, add, 128)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE3_LOOP1_256\n\t"
      /* 7, 5 */
      /* %0 = data+28   %1 = data+160 */
      "addl "     "$4, "      "%3\n\t"
      "STAGE3_LOOP2_256:\n\t"
      SR_BF_k_eq_N8_2R(4, 0, sub, 1, add, 128)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE3_LOOP2_256\n\t"
      "subl "     "$(8*2), "  "%0\n\t"
      "addl "     "$(8*2), "  "%1\n\t"
      SR_BF_k_eq_N8_2R(4, 0, sub, 1, add, 128)
      /* %0 = data      %1 = data+188 */

      "addl "     "$(32*2), " "%1\n\t"
      "addl "     "$2, "      "%3\n\t"
      /* %0 = data      %1 = data+220 */
      "STAGE3_LOOP3_256:\n\t"
      STAGE_OUTPUT___2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT___2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT___2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R__(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R__(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R__(100, 0, add, 1, sub, 64)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE3_LOOP3_256\n\t"
      /* %0 = data+256  %1 = data-36 */

   /* STAGE 4: 16pt-DFTs
      registers used: %0 = data
                      %1 = data+220
                      %2 = IDFT_CONST
                      %3 = 2
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "subl "     "$(256*2), ""%0\n\t"
      "addl "     "$(256*2), ""%1\n\t"
      "addl "     "$2, "      "%3\n\t"
      /* 15, 13 */
      /* 11,  9 */
      /* %0 = data      %1 = data+220 */
      "STAGE4_LOOP1_256:\n\t"
      SR_BF_2R( 8, 0, add, 1, sub, 64, h, l)
      SR_BF_2R( 4, 0, add, 1, sub, 64, h, l)
      SR_BF_2R(52, 0, add, 1, sub, 64, h, l)
      "addl "     "$(8*2), "  "%2\n\t"
      "dec "                  "%3\n\t"
      "jnz "      "STAGE4_LOOP1_256\n\t"
      /* %0 = data+128  %1 = data+92 */
      "addl "     "$(12*2), " "%0\n\t"
      "subl "     "$(12*2), " "%1\n\t"
      /* 7, 5 */
      /* %0 = data+140  %1 = data+80 */
      SR_BF_k_eq_N8_2R( 4, 0, sub, 1, add, 64)
      SR_BF_k_eq_N8_2R( 8, 0, sub, 1, add, 64)
      SR_BF_k_eq_N8_2R(76, 0, add, 1, sub, 64)
      /* 3, 1 */
      /* %0 = data+204  %1 = data+16 */
      SR_BF_k_eq_0_2I( 4, 1, add, 0, sub, 64)
      SR_BF_k_eq_0_2I( 8, 1, add, 0, sub, 64)
      SR_BF_k_eq_0_2I(28, 1, sub, 0, add, 64)
      /* %0 = data+220  %1 = data */

      "addl "     "$(16*2), " "%0\n\t"
      "addl "     "$4, "      "%3\n\t"
      /* %0 = data+236  %1 = data */
      "STAGE4_LOOP2_256:\n\t"
      STAGE_OUTPUT_2R2I( 4, 1, add, 0, sub, 32)
      STAGE_OUTPUT___2I( 4, 1, add, 0, sub, 32)
      STAGE_OUTPUT_2R__( 4, 1, add, 0, sub, 32)
      STAGE_OUTPUT_2R2I(52, 1, add, 0, sub, 32)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE4_LOOP2_256\n\t"
      /* %0 = data-20   %1 = data+256 */

   /* STAGE 5: 32pt-DFTs
      registers used: %0 = data+4
                      %1 = data+232
                      %2 = IDFT_CONST+16
                      %3 = 6
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "addl "     "$(24*2), " "%0\n\t"
      "subl "     "$(24*2), " "%1\n\t"
      "addl "     "$6, "      "%3\n\t"
      /* 31, 29
          :   :
         11,  9 */
      /* %0 = data+4    %1 = data+232 */
      "STAGE5_LOOP1_256:\n\t"
      SR_BF_2R(32, 0, add, 1, sub, 32, l, h)
      "addl "     "$(8*2), "  "%2\n\t"
      "dec "                  "%3\n\t"
      "jnz "      "STAGE5_LOOP1_256\n\t"
      /* 7, 5 */
      /* %0 = data+196  %1 = data+40 */
      SR_BF_k_eq_N8_2R(32, 0, add, 1, sub, 32)
      /* 3, 1 */
      /* %0 = data+228  %1 = data+8 */
      SR_BF_k_eq_0_2I(8, 1, sub, 0, add, 32)
      /* %0 = data+236  %1 = data+0 */

      "addl "     "$(8*2), "  "%0\n\t"
      "addl "     "$8, "      "%3\n\t"
      /* %0 = data+244  %1 = data */
      "STAGE5_LOOP2_256:\n\t"
      STAGE_OUTPUT___2I( 4, 1, add, 0, sub, 16)
      STAGE_OUTPUT_2R__(28, 1, add, 0, sub, 16)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE5_LOOP2_256\n\t"
      /* %0 = data-12   %1 = data+256 */

   /* STAGE 6: 64pt-DFTs
      registers used: %0 = data+4
                      %1 = data+240
                      %2 = IDFT_CONST+64
                      %3 = 14
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "addl "     "$(16*2), " "%0\n\t"
      "subl "     "$(16*2), " "%1\n\t"
      "addl "     "$14, "     "%3\n\t"
      /* 1, 3 */
      /* %0 = data+4    %1 = data+240 */
      SR_BF_k_eq_0_2I(16, 0, add, 1, sub, 16)
      /* 5, 7 */
      /* %0 = data+20   %1 = data+224 */
      SR_BF_k_eq_N8_2R(16, 1, sub, 0, add, 16)
      /*  9, 11
          :   :
         61, 63  */
      /* %0 = data+36   %1 = data+208 */
      "STAGE6_LOOP1_256:\n\t"
      SR_BF_2R(16, 1, sub, 0, add, 16, h, l)
      "addl "     "$(8*2), "  "%2\n\t"
      "dec "                  "%3\n\t"
      "jnz "      "STAGE6_LOOP1_256\n\t"
      /* %0 = data+260  %1 = data-16 */

      "addl "     "$(16*2), " "%1\n\t"
      "subl "     "$(12*2), " "%0\n\t"
      "addl "     "$8, "      "%3\n\t"
      /* %0 = data+248  %1 = data */
      "STAGE6_LOOP2_256:\n\t"
      STAGE_OUTPUT_2R2I(8, 1, add, 0, sub, 8)
      STAGE_OUTPUT_2R2I(8, 0, sub, 1, add, 8)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE6_LOOP2_256\n\t"
      /* %0 = data+120  %1 = data+128 */

   /* Undo Bit-Reverse Order
      registers used: %0 = data+4
                      %1 = data+128
                      %3 = 28
                      %5 = MMX_256_CONST */
   /*  op_code     src         dest */
      "subl "     "$(116*2), ""%0\n\t"
      "addl "     "$28, "     "%3\n\t"
      /* %0 = data+4    %1 = data+128 */
      "SWAP_LOOP_256:\n\t"
      "movq "     "(%0), "    "%%mm0\n\t"
      "movq "     "(%1), "    "%%mm1\n\t"
      "movq "     "%%mm0, "   "(%1)\n\t"
      "movl "     "(%5), "    "%2\n\t"
      "movq "     "%%mm1, "   "(%0)\n\t"
      "movl "     "4(%5), "   "%4\n\t"
      "addl "     "%2, "      "%0\n\t"
      "addl "     "%4, "      "%1\n\t"
      "addl "     "$8, "      "%5\n\t"
      "dec "                  "%3\n\t"
      "jnz "      "SWAP_LOOP_256\n\t"

   /* Post-IDFT Scaling:
      registers used: %0 = data
                      %1 = data+252
                      %3 = 64 */
   /*  op_code     src         dest */
      "addl "     "$64, "     "%3\n\t"
      "subl "     "$(368*2), ""%5\n\t"
      PRE_POST_SCALING_SETUP
      /* %0 = data      %1 = data+252 */
      "POST_LOOP_256:\n\t"
      PRE_POST_SCALING(sub)
      "dec "                  "%3\n\t"
      "jnz "      "POST_LOOP_256\n\t"
      /* %0 = data+256  %1 = data-4 */

      "emms\n\t"
   /* output list */
   :
   /* input list */
   : "q" (data),
     "q" (data+252),
     "q" (MMX_IDFT_CONST+256),
     "r" (64),
     "r" (0x5A815A81),
     "q" (MMX_256_CONST)
   );
}

#undef STAGE1
#undef STAGE_OUTPUT_2R2I
#undef STAGE_OUTPUT___2I
#undef STAGE_OUTPUT_2R__
#undef SR_BF_k_eq_0_2I
#undef SR_BF_k_eq_N8_2R
#undef SR_BF_2R
#undef PRE_POST_SCALING_SETUP
#undef PRE_POST_SCALING

