/***************************************************************************
                          PMT_SECTION.h  -  description
                             -------------------
   Stuff needed for handling Program Map Table data.
                             -------------------
    begin                : Sat Aug 21 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PMT_SECTION_H__
#define __PMT_SECTION_H__

#include "common/typedefs.h"

#ifndef PMT_MAX_PROG_DESC
# define PMT_MAX_PROG_DESC 4
#endif
#ifndef PMT_MAX_ELEM_DESC
# define PMT_MAX_ELEM_DESC 6
#endif


#define DESC_TAG_VIDEO_STREAM         0x02
#define DESC_TAG_AUDIO_STREAM         0x03
#define DESC_TAG_HIERARCHY            0x04
#define DESC_TAG_REGISTRATION         0x05
#define DESC_TAG_DATA_ALIGN           0x06
#define DESC_TAG_TGT_BACKGROUND_GRID  0x07
#define DESC_TAG_VIDEO_WIN            0x08
#define DESC_TAG_CA                   0x09
#define DESC_TAG_LANGUAGE             0x0A
#define DESC_TAG_SYS_CLK              0x0B
#define DESC_TAG_MULTI_BUF_UTIL       0x0C
#define DESC_TAG_COPYRIGHT            0x0D
#define DESC_TAG_MAX_BITRATE          0x0E
#define DESC_TAG_PRIVATE_DATA         0x0F
#define DESC_TAG_SMOOTH_BUF           0x10
#define DESC_TAG_STD                  0x11
#define DESC_TAG_IBP                  0x12
#define DESC_TAG_ATSC_AC3_AUDIO       0x81

#define DESC_TAG_RESERVED(ptr_to_PMT_DESC) \
   ( (ptr_to_PMT_DESC->tag < 0x02)|| \
     ((0x12 < ptr_to_PMT_DESC->tag)&&(ptr_to_PMT_DESC->tag < 0x40)) )
#define DESC_TAG_USR_DEFINED(ptr_to_PMT_DESC) \
   (ptr_to_PMT_DESC->tag > 0x3F)
#define DESC_TAG_ATSC(ptr_to_PMT_DESC) \
   ( (0x3F < ptr_to_PMT_DESC->tag)&&(ptr_to_PMT_DESC->tag < 0xB0) )

   
#define STREAM_TYPE_MPEG1_VIDEO        0x01
#define STREAM_TYPE_MPEG2_VIDEO        0x02
#define STREAM_TYPE_MPEG1_AUDIO        0x03
#define STREAM_TYPE_MPEG2_AUDIO        0x04
#define STREAM_TYPE_PRIVATE            0x05
#define STREAM_TYPE_PES_PRIVATE        0x06
#define STREAM_TYPE_ISO_13522_MHEG     0x07
#define STREAM_TYPE_ISO_13818_1_DSMCC  0x08
#define STREAM_TYPE_ITUT_H2221         0x09
#define STREAM_TYPE_ISO_13818_6_A      0x0A
#define STREAM_TYPE_ISO_13818_6_B      0x0B
#define STREAM_TYPE_ISO_13818_6_C      0x0C
#define STREAM_TYPE_ISO_13818_6_D      0x0D
#define STREAM_TYPE_ISO_13818_1_AUX    0x0E
#define STREAM_TYPE_ATSC_AC3_AUDIO     0x81
#define STREAM_TYPE_WMV_DATA           0xB0  /* this is just a temporary hack */
/*                     <0x00> reserved
                  <0x0F-0x7F> reserved
                  <0x80-0xFF> user defined private streams
                  <0x80-0xAF> ATSC private streams
                       <0x81> ATSC AC3 audio stream */

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8 desc_tag,         /* 8: Identifies this descriptor. Use the DESC_TAG*
                                macros to correctly identify the descriptors. */
      desc_length,      /* 8: Number of bytes, in the following data array.
               <254> Max value */
      data[2];
} PMT_DESC;

typedef struct {
   u8        stream_type,     /*  8: Indicates the type of data contained
                                       in the elementary_PID stream. Use the
                                       STREAM_TYPE* macros to correctly
                                       identify the streams. */
             desc_size;       /*     Number of valid pointers in the desc
                                       array. */
   u16       elementary_PID;  /* 13: The PID of the TS packets carrying the
                                       elementary stream. */
   PMT_DESC *desc[PMT_MAX_ELEM_DESC];
} ELEM_STREAM;

typedef struct {
   u8           desc_size,    /*     Number of valid pointers in the desc
                                       array. */
                elem_size;    /*     Number of valid pointers in the elem array. */
   u16          PCR_PID;      /* 13: The PID of the TS packets carrying
                                       the PCR for this program.
                        <0x1FFF> Used when there is no PCR. */
   PMT_DESC    *desc[PMT_MAX_PROG_DESC];
   ELEM_STREAM *elem[1];
} PMT_SECTION;

#ifdef __cplusplus
};
#endif

#endif
