/***************************************************************************
                          mpeg_ps.c  -  description
                             -------------------
   Functions for handling mpeg packet streams.
                             -------------------
    begin                : Fri Mar 22 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define MPEG_PS_TRACE 1*/
#ifdef MPEG_PS_TRACE
# include <stdio.h>
#endif
#include <stdlib.h>
#include <string.h>

#include "common/Timer.h"
#include "MPEG2_SYS.h"
#include "mpeg_es.h"
#include "mpeg_ps.h"
#include "mpeg_ps_priv.h"
#include "commands.h"
#include "detect_mpeg.h"
#include "TS_PKT.h"
#include "PMT_SECTION.h"


/* A player function to be used with the MPEG2_SYS_Play() function.
     This function can handle MPEG packet streams. */
void * THREAD_FUNC_TYPE MPEG_PS_Play(void *arg)
{
   PES_PKT        pes;
   const u8      *ptr;
   u32            buf_size, i;
   CLK_VAL       *pos;
   MPEG2_SYS_ARG *sys_arg = (MPEG2_SYS_ARG *)arg;
   
   memset(&pes, 0, sizeof(PES_PKT));
   if (sys_arg->si == VIDEO_DECODERS) {   /* ignore TS PSI modules */
      sys_arg->fi = sys_arg->si + AUDIO_DECODERS;
      pos         = &(sys.data.a_pos);
   }
   else
      pos         = &(sys.data.v_pos);

/*   printf("%s> start\n", (sys_arg->si) ? "AUDIO" : "VIDEO");*/

   /* check if there is anything to do */
   buf_size = 0;
   lock_MUTEX(sys.m, 1);
   sys_arg->pos = 0;
   while (sys.data.command != CMD_STOP) {
      for (i = sys_arg->si; i < sys_arg->fi; i++) {
         if (sys.PID[i] < TS_PKT_NUL_ID)
            goto MPEG_PS_Play_START_WORKING;
      }
      if (sys_arg->bytes_read > buf_size) {
         sys_arg->buf += sys_arg->bytes_read - buf_size;
         if (sys_arg->buf > sys.last_buf)
            sys_arg->buf = sys.buf;
      }
      else if (!sys_arg->buf) {
         sys_arg->buf++;
         broadcast_CONDVAR(sys.cv); /* acknowledge the discontinuity notice */
      }
      sys_arg->bytes_read = 
      buf_size            = 1;
      wait_CONDVAR(sys.cv, sys.m);
      pos[0] = pos[2] = 0;
   }
   goto MPEG_PS_Play_SHUTDOWN;
MPEG_PS_Play_START_WORKING:
   if ( (sys_arg->bytes_read)&&(buf_size) )
      sys_arg->bytes_read -= buf_size;
   unlock_MUTEX(sys.m);
/*
   if (sys.data.command != CMD_STOP)
      printf("%s> working\n", (sys_arg->si) ? "AUDIO" : "VIDEO");
*/
   while (sys.data.command != CMD_STOP) {
      lock_MUTEX(sys.m, 1);
      sys_arg->pos = *pos;
      while (!sys_arg->bytes_read) {
         if (!sys_arg->buf) {
            sys_arg->buf++;
            pes.payload = NULL;
            broadcast_CONDVAR(sys.cv); /* acknowledge the discontinuity notice */
         }
         wait_CONDVAR(sys.cv, sys.m);
         if (sys.data.command == CMD_STOP)
            goto MPEG_PS_Play_SHUTDOWN;   /* shutdown */
      }
      if (sys_arg->bytes_read > BUFFER_SIZE) {
         buf_size             = BUFFER_SIZE;
         sys_arg->bytes_read -= BUFFER_SIZE;
      }
      else {
         buf_size             = sys_arg->bytes_read;
         sys_arg->bytes_read  = 0;  /* ask MPEG2_SYS_Play() thread for more data */
      }
      ptr = sys_arg->buf;
      if (sys_arg->buf >= sys.last_buf)
         sys_arg->buf  = sys.buf;
      else
         sys_arg->buf += BUFFER_SIZE;
      broadcast_CONDVAR(sys.cv);
      unlock_MUTEX(sys.m);

      if (pes.payload) {
         pes.payload = ptr;
         if ((u32)pes.packet_length < buf_size) {
            pes.payload_len16 = pes.packet_length;
            ptr              += pes.packet_length;
         }
         else
            pes.payload_len16 = (u16)buf_size;
      }

      while (buf_size) {
         while ( (!pes.payload)&&(buf_size) ) {
            ptr = PS_find_packet(ptr, &buf_size, &pes);
         }
   
         if (pes.payload) {
            pes.packet_length -=      pes.payload_len16;
            buf_size          -= (u32)pes.payload_len16;

            if (pes.flags1) {
/*********** START TO DO SOMETHING WITH THE PES PACKET ***********/
               for (i = sys_arg->si; i < sys_arg->fi; i++) {
                  if (sys.PID[i] == pes.stream_id2) {
                     sys.psi_pes[i].process_pkt(sys.psi_pes[i].common.mem,
                                                &pes);
                     break;
                  }
               }
/*********** STOP DOING SOMETHING WITH THE PES PACKET ************/
            }

            if (!pes.packet_length) {
               pes.payload = NULL;
#ifdef _RECORD_STATS_
               sys_arg->pkt_cnt++;
#endif
            }
            else
               pes.payload = (const u8 *)1;
         }
      }
   }

   lock_MUTEX(sys.m, 1);
MPEG_PS_Play_SHUTDOWN:
/*   i = 0;*/
   do {
/*      i++;*/
      sys_arg->bytes_read = 0;
      broadcast_CONDVAR(sys.cv);
      wait_CONDVAR(sys.cv, sys.m);
   } while (sys_arg->bytes_read != 0xFFFFFFFF); /* wait for everyone to stop */
   unlock_MUTEX(sys.m);
/*   printf("%s> end after %u attempts\n", (sys_arg->si) ? "AUDIO" : "VIDEO", i);*/
   exit_THREAD(NULL);
   return NULL;
}

/* Creates a program, using the first audio and video streams
     found in the given MPEG packet stream.
      INPUT: fp - Points to the FILE with the MPEG packet stream.
      RETURN: Program number, of newly created program, is stored
                in the global struct (sys.program_number).
              Success - A non-zero value
              Failure - 0 */
void MPEG_PS_Create_Program(FILE_PTR fp)
{
   FILE_POS          fp_pos;
   int               mod;
   PS_CREATE_PROGRAM arg      = {TS_PKT_NUL_ID, TS_PKT_NUL_ID};
   u8                vid_type = 0,
                     aud_type = 0;

   FILE_GETPOS(fp, fp_pos);
   MPEG_PS_decode(fp, (fp_pos+CMD_SEARCH_SIZE), PS_create_program, &arg);
   FILE_SETPOS(fp, fp_pos);

   /* find correct stream type values */
   if (arg.vid_PID < TS_PKT_NUL_ID) {
      vid_type = MPEG_PS_stream_type(fp, arg.vid_PID);
      if (!vid_type)
         arg.vid_PID = TS_PKT_NUL_ID;
      else {
         mod = MPEG_ES_find_module(vid_type);
         if (mod < 0)
            arg.vid_PID = TS_PKT_NUL_ID;
         else
            vid_type    = (u8)mod;
      }
   }
   if (arg.aud_PID < TS_PKT_NUL_ID) {
      aud_type = MPEG_PS_stream_type(fp, arg.aud_PID);
      if (!aud_type)
         arg.aud_PID = TS_PKT_NUL_ID;
      else {
         mod = MPEG_ES_find_module(aud_type);
         if (mod < 0)
            arg.aud_PID = TS_PKT_NUL_ID;
         else
            aud_type    = (u8)mod;
      }
   }

   if ( (arg.vid_PID >= TS_PKT_NUL_ID)&&
        (arg.aud_PID >= TS_PKT_NUL_ID) )
      sys.program_number = 0;
   else {
      sys.program_number = 1;
      
      if (arg.vid_PID < TS_PKT_NUL_ID)
         sys.PID[vid_type] = arg.vid_PID;

      if (arg.aud_PID < TS_PKT_NUL_ID)
         sys.PID[aud_type] = arg.aud_PID;
   }
}

/* Gets the stream ids and types from the packet stream.
      INPUT: fp   - Points to the FILE with the MPEG packet
                      stream.
             size - Number of elements in the PID and stream_type
                      arrays.
      OUTPUT: PID         - Array of stream ids found in the packet
                              stream.
              stream_type - Array of STREAM_TYPE_* values 
                              corresponding to the values in the
                              PID array.
      RETURN: The number of PIDs found. */
u32 MPEG_PS_get_PIDs(FILE_PTR fp, u16 *PID, u8 *stream_type, u32 size)
{
   u32         i;
   FILE_POS    fp_pos;
   PS_GET_PIDS arg = {PID, size, 0};

   memset(PID, 0xFF, sizeof(u16)*(size_t)size);
   memset(stream_type, 0, sizeof(u8)*(size_t)size);

   /* get PIDs first */
   FILE_GETPOS(fp, fp_pos);
   MPEG_PS_decode(fp, (fp_pos+CMD_SEARCH_SIZE), PS_get_PIDs, &arg);
   FILE_SETPOS(fp, fp_pos);

   /* now get the stream_types */
   size = arg.found;
   for (i = 0; i < size; i++) {
      stream_type[i] = MPEG_PS_stream_type(fp, PID[i]);
   }
   return size;
}

/* Tries to determine the type of data contained in the stream with
     the given PID.
      INPUT: fp  - Pointer to an open file that contains MPEG PS data.
             PID - PID of the stream to look for.
      RETURN: Success - One of the STREAM_TYPE_* values defined
                          in PMT_SECTION.h
              Failure - Zero. */
u8 MPEG_PS_stream_type(FILE_PTR fp, u16 PID)
{
   PS_STREAM_TYPE arg;
   FILE_POS       fp_pos;
   
   memset(&arg, 0, sizeof(PS_STREAM_TYPE));
   arg.PID = PID;

   FILE_GETPOS(fp, fp_pos);
   MPEG_PS_decode(fp, (fp_pos+CMD_SEARCH_SIZE), PS_stream_type, &arg);
   FILE_SETPOS(fp, fp_pos);

   return arg.mod_data.stream_type;
}

/* Decodes the MPEG program stream in the given file
      INPUT: fp      - Pointer to an open file that contains MPEG PS 
                         data.
             end_pos - Point at which the decoding will stop. Decoding
                         will stop at an earlier point, if an EOF is
                         encountered.
             func    - Function that will deal with the decoded 
                         PES_PKTs.
             arg     - Pointer to any private data for func. */
void MPEG_PS_decode(FILE_PTR            fp, 
                    FILE_POS            end_pos, 
                    MPEG_PS_decode_func func,
                    void               *arg)
{
   const u8       *ptr;      
   FILE_POS        cur_pos;
   FILE_BYTE_COUNT bytes_read;
   PES_PKT         pes;
   u32             buf_size;
#ifdef MPEG_PS_TRACE
   u32             found   = 0,
                   seq_cnt = 0;
#endif

#ifdef MPEG_PS_TRACE
   FILE_GETPOS(fp, cur_pos);
   printf("***file pos = 0x%X***\n", (u32)cur_pos);
#endif

   memset(&pes, 0, sizeof(PES_PKT));
   do {
      FILE_READ(fp, sys.buf, BUFFER_SIZE, bytes_read);
      if (bytes_read < 1)
         break;
      ptr      = sys.buf;
      buf_size = (u32)bytes_read;
      if (pes.payload) {
         pes.payload = ptr;
         if ((u32)pes.packet_length < buf_size) {
            pes.payload_len16 = pes.packet_length;
            ptr              += pes.packet_length;
         }
         else
            pes.payload_len16 = (u16)buf_size;
      }
      
      do {
         while ( (!pes.payload)&&(buf_size) ) {
#ifdef MPEG_PS_TRACE
            printf("   start find_packet (+0x%X)\n",
                   (u32)(ptr - sys.buf));
            if (pes.state)
               printf("   state=0x%04hX\n", pes.state);
#endif
            ptr = PS_find_packet(ptr, &buf_size, &pes);
#ifdef MPEG_PS_TRACE
            if ( (ptr)&&(pes.payload) ) {
               printf("   PES   (+0x%X -> +0x%X len=%hu)\n", 
                      (u32)(pes.payload - sys.buf),
                      (u32)(pes.payload - sys.buf)+(u32)pes.packet_length,
                      pes.packet_length);
            }
            else if (pes.state)
               printf("   state=0x%04hX\n", pes.state);
#endif
         }

         if (pes.payload) {
            pes.packet_length -=      pes.payload_len16;
            buf_size          -= (u32)pes.payload_len16;

/*********** START TO DO SOMETHING WITH THE PES PACKET ***********/
#ifdef MPEG_PS_TRACE
            while (pes.payload_len16) {
               const u8 *ptr2;
               
               ptr2 = detect_start_code(pes.payload, 
                                        (u32)pes.payload_len16, 
                                        &found);
               if (ptr2) {
                  printf("   0x%02hX  (+0x%X)", 
                         (u16)(*ptr2),
                         (u32)(ptr2 - sys.buf));
                  if ( (*ptr2 == 0xB3)&&(pes.stream_id == 0xE0) ) {
                     seq_cnt++;
                     printf("  seq %u", seq_cnt);
                  }
                  printf("\n");
                  pes.payload_len16 -= (u16)(ptr2 - pes.payload);
                  pes.payload        = ptr2;
                  found              = 0;
               }
               else
                  pes.payload_len16  = 0;
            }

            if (!pes.packet_length) {
               printf("\n");
               found = 0;
            }
#else
            if (!func(&pes, arg)) {
               buf_size = bytes_read = 0;
               break;
            }
#endif
/*********** STOP DOING SOMETHING WITH THE PES PACKET ************/

            pes.payload = (pes.packet_length) ? (const u8 *)1 : NULL;
         }
      } while (buf_size);
      
      FILE_GETPOS(fp, cur_pos);
#ifdef MPEG_PS_TRACE
      printf("***file pos = 0x%X***\n", (u32)cur_pos);
   } while (bytes_read == BUFFER_SIZE);
#else
   } while ( (bytes_read == BUFFER_SIZE)&&(cur_pos < end_pos) );
#endif
}


const u8 *PS_find_packet(const u8 *buf, 
                         u32      *buf_size, 
                         PES_PKT  *pes)
{
   const
   u8  *ptr;
   u32  loc = (u32)(pes->state >> 8);

   pes->state &= 0x00FF;
#define MPEG1_PS_STATE   5
#define MPEG2_PS_STATE   8
#define READ_PTS_STATE  10
#define AC3_STATE       15
#define PAYLOAD_STATE   16
#ifndef MPEG2_SYS_CLK64
# define PTS_SHIFT       2
#else
# define PTS_SHIFT       0
#endif

   while (*buf_size) {
      switch (pes->state) {
      case 0:
         ptr = detect_start_code(buf, *buf_size, &loc);
         if (ptr) {
            if (*ptr >= PES_PROGRAM_STREAM_MAP) {
               if ( ((0xEF < *ptr)||(*ptr < 0xC0))&&
                    (*ptr != PES_ATSC_AC3_AUDIO) ) {
                  pes->flags1 = 0x00;  /* ignore this stream */
#ifdef MPEG_PS_TRACE
                  printf("\nIGNORE   ");
               }
               else
                  printf("\nES STREAM");
               printf(" (0x%02hX) ", (u16)(*ptr));
#else
               }
#endif
               pes->stream_id = *ptr;
               pes->state     = 3;
            }
            else if (*ptr == 0xBA) {
               pes->state++;
#ifdef MPEG_PS_TRACE
               printf("MPEG_PS_PACKET  ");
#endif
            }
#ifdef MPEG_PS_TRACE
            else if (*ptr == 0xB9)
               printf("MPEG_PS_END_CODE");
            else if (*ptr == 0xBB)
               printf("MPEG_SYS_HEADER ");
            else
               printf("0x%02hX             ", (u16)(*ptr));
            printf(" (+0x%X)\n", (u32)(ptr - sys.buf));
#endif
            loc        = 0;
            *buf_size -= (u32)(ptr - buf);
         }
         else
            *buf_size  = 0;
         buf = ptr;
         break;
      
      case 1:  /* determine stream type */
#ifdef MPEG_PS_TRACE
         printf("  STREAM TYPE: ");
#endif
         if ((*buf & 0xC0) == 0x40) {
#ifdef MPEG_PS_TRACE
               printf("MPEG2\n");
#endif
            pes->flags1 = 0x80;
         }
         else if ((*buf & 0xF0) == 0x20) {
#ifdef MPEG_PS_TRACE
               printf("MPEG1\n");
#endif
            pes->flags1 = 0x40;
         }
         else {
#ifdef MPEG_PS_TRACE
               printf("UNKNOWN\n");
#endif
            pes->flags1 = 0x00;
         }
         pes->state = 0;
         break;

      case 2:  /* ignore bytes, before changing state */
         if (--loc == 0)
            pes->state = pes->stream_id2;
         break;
         
      case 3:  /* read packet length */
         pes->packet_length  = (u16)*buf << 8;
         pes->state++;
         break;
      case 4:
         pes->packet_length |= (u16)*buf;
#ifdef MPEG_PS_TRACE
         printf("  packet_length=%hu (0x%04hX)\n", 
                pes->packet_length, pes->packet_length);
#endif

         if (pes->flags1 == 0x40)         /* MPEG1 PS */
            pes->state++;
         else if (pes->flags1 == 0x80) {  /* MPEG2 PS */
            /* skip PES header flags */
            pes->packet_length -= 3;
            pes->stream_id2     = MPEG2_PS_STATE;
            pes->state          = 2;
            loc                 = 1;
         }
         else
            pes->state          = PAYLOAD_STATE;
         break;

/*****START MPEG1 PS*****/
      case MPEG1_PS_STATE:    /* skip stuffing bytes */
         if (*buf == 0xFF) {
            pes->packet_length--;
            break;
         }
      
      case MPEG1_PS_STATE+1:  /* look for buffer parameters */
         if ((*buf & 0xC0) == 0x40) {
            pes->packet_length -=  2;
            pes->stream_id2     = MPEG1_PS_STATE + 2;
            pes->state          = 2;
            loc                 = 1;
            break;
         }

      case MPEG1_PS_STATE+2:  /* look for PTS and DTS */
         if ((*buf & 0xE0) == 0x20) {
            pes->packet_length -= 5;
            pes->header_length  = (*buf & 0x10) ? 5 : 0;
            pes->flags2         = 0x80;   /* set PTS flag */
            goto find_packet_READ_PTS;
         }
         else {
            pes->packet_length--;
            pes->flags2 = 0;  /* clear PTS flag */
            pes->state  = PAYLOAD_STATE;
            break;
         }
/*****END MPEG1 PS*****/

/*****START MPEG2 PS*****/
      case MPEG2_PS_STATE:    /* read PES flags */
         pes->flags2 = *buf;
         pes->state++;
         break;
      
      case MPEG2_PS_STATE+1:  /* read header length */
         pes->header_length = *buf;
         if (PES_PTS_FLAG(pes)) {
            pes->header_length -= 5;
            pes->packet_length -= 5;
            pes->state++;
         }
         else if (pes->flags2 & 0x40)
            pes->state = PAYLOAD_STATE + 50;
         else
            goto find_packet_READ_PTS_DONE;
         break;
/*****END MPEG2 PS*****/

      case READ_PTS_STATE:
find_packet_READ_PTS:
         pes->PTS   = (CLK_VAL)(*buf & 0x0E) << (29 - PTS_SHIFT);
         pes->state = READ_PTS_STATE + 1;
         break;
      case READ_PTS_STATE+1:
         pes->PTS  |= (CLK_VAL) *buf         << (22 - PTS_SHIFT);
         pes->state++;
         break;
      case READ_PTS_STATE+2:
         pes->PTS  |= (CLK_VAL)(*buf & 0xFE) << (14 - PTS_SHIFT);
         pes->state++;
         break;
      case READ_PTS_STATE+3:
         pes->PTS  |= (CLK_VAL) *buf         << ( 7 - PTS_SHIFT);
         pes->state++;
         break;
      case READ_PTS_STATE+4:
         pes->PTS  |= (CLK_VAL)(*buf         >> ( 1 + PTS_SHIFT));
#ifdef MPEG2_SYS_CLK64
         pes->PTS  *= 300;
#endif
#ifdef MPEG_PS_TRACE
         printf("  PTS=%d\n", pes->PTS);
#endif
find_packet_READ_PTS_DONE:
         if (pes->header_length) {
            pes->packet_length -= (u16)pes->header_length;
            pes->stream_id2     = PAYLOAD_STATE;
            pes->state          = 2;   /* skip the rest of the header extension */
            loc                 = (u32)pes->header_length;
         }
         else
            pes->state          = PAYLOAD_STATE;
         break;

/*****START AC3*****/
      case AC3_STATE:
#ifdef MPEG_PS_TRACE
         printf(" 0x%02hX", (u16)*buf);
#endif
         if (--loc == 0) {
#ifdef MPEG_PS_TRACE
/*            printf(" (0x%02hX 0x%02hX)\n", (u16)*(buf+1), (u16)*(buf+2));*/
            printf("  ID2=0x%04hX\n", pes->stream_id2);
#endif
            pes->state = PAYLOAD_STATE + 1;
         }
         break;
/*****END AC3*****/

      case PAYLOAD_STATE:  /* finally found the payload! */
         pes->stream_id2 = (u16)pes->stream_id;
         if (pes->stream_id == PES_ATSC_AC3_AUDIO) {
/*
            AC3 Multiplex Bytes:
            [0] = 0x80 + stream_num
            [1] = num_syncwords
            [2] = high_byte(first_syncword + 1)
            [3] =  low_byte(first_syncword + 1)
*/
            pes->stream_id2    |= (u16)(*buf & 0x7F) << 8;
            pes->state          = AC3_STATE;
            pes->packet_length -= 4;
            loc                 = 3;   /* skip the AC3 multiplex bytes */
#ifdef MPEG_PS_TRACE
            printf("  AC3=0x%02hX", (u16)*buf);
#endif
            break;
         }
      case PAYLOAD_STATE+1:
         pes->payload = buf;         
         if (*buf_size <= (u32)pes->packet_length)
            pes->payload_len16 = (u16)*buf_size;
         else {
            pes->payload_len16 = pes->packet_length;
            buf               += pes->packet_length;
         }
      default:
         pes->state = 0;
         return buf;
      }
   
      if (*buf_size) {
         (*buf_size)--;
         buf++;
      }
      else
         break;
   }

   pes->state |= (u16)(loc << 8);
   return buf;
#undef MPEG1_PS_STATE
#undef MPEG2_PS_STATE
#undef READ_PTS_STATE
#undef AC3_STATE
#undef PAYLOAD_STATE
#undef PTS_SHIFT
}

int PS_create_program(PES_PKT *pes, void *arg)
{
   PS_CREATE_PROGRAM *data = (PS_CREATE_PROGRAM *)arg;

   if ( (pes->stream_id2 != data->vid_PID)&&
        (pes->stream_id2 != data->aud_PID) ) {
      if (data->vid_PID >= TS_PKT_NUL_ID) {
         if ( (0xDF < pes->stream_id)&&(pes->stream_id < 0xF0) )
            data->vid_PID = pes->stream_id2;
      }

      if (data->aud_PID >= TS_PKT_NUL_ID) {
         if ( ((0xBF < pes->stream_id)&&(pes->stream_id < 0xE0))||
              (pes->stream_id == PES_ATSC_AC3_AUDIO) )
            data->aud_PID = pes->stream_id2;
      }
   }

   if ( (data->vid_PID < TS_PKT_NUL_ID)&&
        (data->aud_PID < TS_PKT_NUL_ID) )
      return 0;   /* we're done! */
   return 1;
}

int PS_get_PIDs(PES_PKT *pes, void *arg)
{
   u32          i;
   PS_GET_PIDS *data = (PS_GET_PIDS *)arg;

   if (pes->flags1) {
      for (i = 0; i < data->size; i++) {
         if (data->PID[i] == pes->stream_id2)
            break;
      }
   
      if (i >= data->size) {
         /* found a new PID */
         data->PID[data->found] = pes->stream_id2;
         data->found++;
         if (data->found >= data->size)
            return 0;   /* we're done! */
      }
   }
   return 1;
}

int PS_stream_type(PES_PKT *pes, void *arg)
{
   PS_STREAM_TYPE *data = (PS_STREAM_TYPE *)arg;
   
   if (pes->stream_id2 == data->PID) {
      DETECT_MODULE_process_pkt(&(data->mod_data), pes);
      if (data->mod_data.stream_type)
         return 0;   /* we're done! */
   }
   return 1;
}

