/***************************************************************************
                         mpeg_dvd.h  -  description
                             -------------------
   API to provide DVD access to this application.  This API mimics the
    one found in mpeg_ps.h
                             -------------------
    begin                : Mon Feb 6 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __MPEG_DVD_H__
#define __MPEG_DVD_H__

#include "common/typedefs.h"
#include "dvd_input.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Gets information about the DVD.
      INPUT: dvd  - DVD pointer.
             size - Number of elements in the chapters array.
      OUTPUT: title    - Title of the DVD.
              chapters - Array of chapters, for each title, on the
                           DVD.
      RETURN: The number of titles found. */
u32  MPEG_DVD_get_info(DVD_PTR dvd, const char **title, u16 *chapters, u32 size);

/* Creates a program, using the first audio and video streams
     found in the DVD.
      INPUT: dvd - DVD pointer.
      RETURN: Program number, of newly created program, is stored
                in the global struct (sys.program_number).
              Success - A non-zero value
              Failure - 0 */
void MPEG_DVD_Create_Program(DVD_PTR dvd);

/* Gets the stream ids and types from the DVD.
      INPUT: dvd  - DVD pointer.
             size - Number of elements in the PID and stream_type
                      arrays.
      OUTPUT: PID         - Array of stream ids found in the packet
                              stream.
              stream_type - Array of STREAM_TYPE_* values 
                              corresponding to the values in the
                              PID array.
      RETURN: The number of PIDs found. */
u32  MPEG_DVD_get_PIDs(DVD_PTR dvd, u16 *PID, u8 *stream_type, u32 size);

/* Tries to determine the type of data contained in the stream with
     the given PID.
      INPUT: dvd - DVD pointer.
             PID - PID of the stream to look for.
      RETURN: Success - One of the STREAM_TYPE_* values defined
                          in PMT_SECTION.h
              Failure - Zero. */
u8   MPEG_DVD_stream_type(DVD_PTR dvd, u16 PID);

/* Decodes the MPEG program stream from the DVD.
      INPUT: dvd     - DVD pointer.
             end_pos - Point at which the decoding will stop. Decoding
                         will stop at an earlier point, if an EOF is
                         encountered.
             func    - Function that will deal with the decoded 
                         PES_PKTs.
             arg     - Pointer to any private data for func. */
void MPEG_DVD_decode(DVD_PTR             dvd, 
                     DVD_POS             end_pos, 
                     MPEG_PS_decode_func func,
                     void               *arg);

/* Plays the DVD.
      INPUT: dvd - DVD pointer.
      RETURN: Success - A non-zero value
              Failure - 0 */
int  MPEG_DVD_Play(DVD_PTR dvd, u32 start_chapter);

#ifdef __cplusplus
};
#endif

#endif

