/***************************************************************************
                          PIC_HEADER.c  -  description
                             -------------------
   Stuff needed for handling MPEG2 picture and gop headers.
                             -------------------
    begin                : Tue Sep 28 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef _M2V_TRACE_
# include <stdio.h>
#endif

#include "PIC_HEADER.h"

#ifdef _M2V_TRACE_
static void display_PIC_HEADER(PIC_HEADER *hdr);
#endif

#define EXTRA_BYTE(n, m) \
   case n: \
      if (!(*ptr & m)) { \
         goto PIC_HEADER_read_FINISH; \
      } \
      ptr++; \
      if (!(--data_len)) { \
         *cur_pos = n+1; \
         break; \
      }


/* Reads the gop header in data_stream.
      INPUT: data_stream - Points to the data stream holding the gop
                             header. This stream does not have to
                             contain the entire header. Repeated calls to
                             this function can be used to read the header
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 GOP_HEADER_read(const u8         *data_stream,
                          u32         data_len,
                          PIC_HEADER *hdr,
                          u32        *cur_pos,
                          u16        *code)
{
   switch (*cur_pos) {
   case 0:
      hdr->type &= 0xF8;
      hdr->type |= (*data_stream & 0x80) ? 0x04 : 0x00;
      if (data_len >= 4) {
         data_stream += 3;
         goto GOP_HEADER_read_FINISH;
      }

      *cur_pos = data_len;
      break;
   case 1:
   case 2:
      if (data_len <= (3 - *cur_pos)) {
         *cur_pos += data_len;
         break;
      }
      data_stream += 3 - *cur_pos;
   case 3:
GOP_HEADER_read_FINISH:
      hdr->type |= (*data_stream & 0x60) >> 5;
#ifdef _M2V_TRACE_
         printf("GOP_HEADER> drop_frame = %c"
                          "  closed_gop = %c"
                          "  broken_link = %c\n",
                GOP_HEADER_DROP_FRAME_FLAG(hdr) ?  'Y' : 'N',
                GOP_HEADER_CLOSED_GOP_FLAG(hdr) ?  'Y' : 'N',
                GOP_HEADER_BROKEN_LINK_FLAG(hdr) ? 'Y' : 'N');
#endif
      /* FINISHED READING THE GOP HEADER */
      data_len = 4 - *cur_pos;
      *cur_pos = 0;
      *code   |= 0x8000;   /* have to look for next start code */
      break;
   default:
/*      printf("GOP_HEADER_read: ERROR illegal pos %u\n", *cur_pos); */
      *cur_pos = 0;
      *code    = 0x0FFF;
      return 0;
   }

   return data_len;
}

/* Reads the picture header in data_stream.
      INPUT: data_stream - Points to the data stream holding the picture
                             header. This stream does not have to
                             contain the entire header. Repeated calls to
                             this function can be used to read the header
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 PIC_HEADER_read(const u8         *data_stream,
                          u32         data_len,
                          PIC_HEADER *hdr,
                          u32        *cur_pos,
                          u16        *code)
{
   const u8 *ptr = data_stream;

   switch (*cur_pos) {
   case 0:
      hdr->temporal_reference  = (u16)(*ptr) << 2;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 1;
         break;
      }
   case 1:
      hdr->temporal_reference |= (u16)(*ptr >> 6);
      hdr->type               &= 0x8F;
      hdr->type               |= (*ptr & 0x38) << 1;
      hdr->vbv_delay           = (u16)(*ptr & 0x07) << 13;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 2;
         break;
      }
   case 2:
      hdr->vbv_delay          |= (u16)(*ptr) << 5;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 3;
         break;
      }
   case 3:
      hdr->vbv_delay          |= (u16)(*ptr) >> 3;
/*
      if (PIC_HEADER_CODING_TYPE(hdr) == CODING_TYPE_I_FRAME)
         goto PIC_HEADER_read_EXTRA_4;

      ptr++;
      if (PIC_HEADER_CODING_TYPE(hdr) == CODING_TYPE_P_FRAME) {
         if (!(--data_len)) {
            *cur_pos = 9;
            break;
         }
         goto PIC_HEADER_read_EXTRA_9;
      }
      else if (!(--data_len)) {
         *cur_pos = 4;
         break;
      }

PIC_HEADER_read_EXTRA_4:
   EXTRA_BYTE(4, 0x04)
   EXTRA_BYTE(5, 0x02)
   EXTRA_BYTE(6, 0x01)
   case 7:
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 8;
         break;
      }
   EXTRA_BYTE(8, 0x80)
PIC_HEADER_read_EXTRA_9:
   EXTRA_BYTE(9, 0x40)
   EXTRA_BYTE(10, 0x20)
   EXTRA_BYTE(11, 0x10)
   case 12:
      if (*ptr & 0x08) {
         ptr++;
         if (!(--data_len)) {
            *cur_pos = 4;
            break;
         }
         goto PIC_HEADER_read_EXTRA_4;
      }

PIC_HEADER_read_FINISH:
*/
#ifdef _M2V_TRACE_
      {
         char type;

         switch (PIC_HEADER_CODING_TYPE(hdr)) {
         case CODING_TYPE_I_FRAME:
            type = 'I';
            break;
         case CODING_TYPE_P_FRAME:
            type = 'P';
            break;
         case CODING_TYPE_B_FRAME:
            type = 'B';
            break;
         default:
            type = '-';
         }
         printf("PIC_HEADER_read: coding_type = %c\n", type);
      }
#endif
      /* FINISHED READING THE PIC HEADER */
      ptr++;
      *cur_pos = 0;
      *code   |= 0x8000;   /* have to look for next start code */
      break;

   default:
/*      printf("PIC_HEADER_read: ERROR illegal pos %u\n", *cur_pos); */
      *cur_pos = 0;
      *code    = 0x0FFF;
      return 0;
   }

   return (u32)(ptr - data_stream);
}

/* Reads the picture coding extension in data_stream.
      INPUT: data_stream - Points to the data stream holding the picture
                             coding extension. This stream does not have to
                             contain the entire extension. Repeated calls to
                             this function can be used to read the extension
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER that will hold the data.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 PIC_CODING_EXT_read(const u8         *data_stream,
                              u32         data_len,
                              PIC_HEADER *hdr,
                              u32        *cur_pos,
                              u16        *code)
{
   const u8 *ptr = data_stream;

   switch (*cur_pos) {
   case 0:
      hdr->f_code[0][0] = *ptr & 0x0F;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 1;
         break;
      }
   case 1:
      hdr->f_code[0][1] = *ptr >> 4;
      hdr->f_code[1][0] = *ptr & 0x0F;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 2;
         break;
      }
   case 2:
      hdr->f_code[1][1] = *ptr >> 4;
/*
      if (PIC_HEADER_CODING_TYPE(hdr) != CODING_TYPE_I_FRAME) {
         hdr->type &= 0x8F;
         if ( (hdr->f_code[1][1] == 0x0F)&&(hdr->f_code[1][0] == 0x0F) ) {
            * I or P frame *
            if ( (hdr->f_code[0][1] == 0x0F)&&(hdr->f_code[0][0] == 0x0F) )
               hdr->type |= CODING_TYPE_I_FRAME;
            else * if ( (hdr->f_code[0][1])&&(hdr->f_code[0][0]) ) *
               hdr->type |= CODING_TYPE_P_FRAME;
         }
         else * if ( (hdr->f_code[1][1])&&(hdr->f_code[1][0]) ) * {
            * B frame *
            hdr->type |= CODING_TYPE_B_FRAME;
         }
      }
*/
      hdr->intra_dc_precision  = (*ptr & 0x0C) >> 2;
      hdr->intra_dc_precision += 8;
      hdr->pic_struct          = *ptr & 0x03;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 3;
         break;
      }
   case 3:
      hdr->pic_struct |= *ptr & 0x80;
      hdr->flags       = *ptr & 0x7F;
      ptr++;
      if (!(--data_len)) {
         *cur_pos = 4;
         break;
      }
   case 4:
      hdr->flags      |= *ptr & 0x80;
      ptr++;
#ifdef _M2V_TRACE_
         display_PIC_HEADER(hdr);
#endif
/*      {
         char type;

         switch (PIC_HEADER_CODING_TYPE(hdr)) {
         case CODING_TYPE_I_FRAME:
            printf("\n");
            type = 'I';
            break;
         case CODING_TYPE_P_FRAME:
            type = 'P';
            break;
         case CODING_TYPE_B_FRAME:
            type = 'B';
            break;
         default:
            type = '-';
         }
         printf("%hu%c ", hdr->temporal_reference, type);
      } */

      /* FINISHED READING PIC CODING EXT */
      *cur_pos = 0;
      *code   |= 0x8000;   /* have to look for next start code */
      break;

   default:
/*      printf("PIC_CODING_EXT_read: ERROR illegal pos %u\n", *cur_pos); */
      *cur_pos = 0;
      *code    = 0x0FFF;
      return 0;
   }

   return (u32)(ptr - data_stream);
}

#ifdef _M2V_TRACE_
static void display_PIC_HEADER(PIC_HEADER *hdr)
{
   const char *coding_type, *pic_struct;

   switch (PIC_HEADER_CODING_TYPE(hdr)) {
   case CODING_TYPE_I_FRAME:
      coding_type = "I Frame";
      break;
   case CODING_TYPE_P_FRAME:
      coding_type = "P Frame";
      break;
   case CODING_TYPE_B_FRAME:
      coding_type = "B Frame";
      break;
   default:
      coding_type = "Reserved";
   }

   switch (PIC_HEADER_STRUCTURE(hdr)) {
   case STRUCTURE_RESERVED:
      pic_struct = "Reserved";
      break;
   case STRUCTURE_TOP_FIELD:
      pic_struct = "Top Field";
      break;
   case STRUCTURE_BOT_FIELD:
      pic_struct = "Bottom Field";
      break;
   case STRUCTURE_FULL_FRAME:
      pic_struct = "Full Frame";
   }
   
   printf("PIC_HEADER> temporal_reference = %hu\n"
          "            coding_type        = %s\n"
          "            vbv_delay          = %hu\n"
          "            intra_dc_precision = %hu\n"
          "            pic_struct         = %s\n"
          "            f_code             = %hu %hu %hu %hu\n"
          "           ",
          hdr->temporal_reference,
          coding_type,
          hdr->vbv_delay,
          (u16)hdr->intra_dc_precision,
          pic_struct,
          hdr->f_code[0][0], hdr->f_code[0][1],
          hdr->f_code[1][0], hdr->f_code[1][1]);
   if (PIC_HEADER_TOP_FIELD_FIRST(hdr))
      printf(" top_field_first");
   if (PIC_HEADER_FRAME_PRED_DCT(hdr))
      printf(" frame_pred_dct");
   if (PIC_HEADER_CONCEALMENT_VECTORS(hdr))
      printf(" concealment_vectors");
   if (PIC_HEADER_Q_SCALE_TYPE(hdr))
      printf(" q_scale_type");
   if (PIC_HEADER_INTRA_VLC_FORMAT(hdr))
      printf(" intra_vlc_format");
   if (PIC_HEADER_ALT_SCAN(hdr))
      printf(" alt_scan");
   if (PIC_HEADER_REPEAT_FIRST_FIELD(hdr))
      printf(" repeat_first_field");
   if (PIC_HEADER_CHROMA_420_TYPE(hdr))
      printf(" chroma_420_type");
   if (PIC_HEADER_PROGRESSIVE_FRAME(hdr))
      printf(" progressive_frame");
   printf("\n");
}
#endif

#undef EXTRA_BYTE

