/***************************************************************************
                          SLICE.h  -  description
                             -------------------
   Stuff needed for handling MPEG2 slices.
                             -------------------
    begin                : Tue Oct 5 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __SLICE_H__
#define __SLICE_H__

#include "common/typedefs.h"
#include "MPEG2_SYS.h"
#include "SEQ_HEADER.h"
#include "PIC_HEADER.h"


#define FRAME_420_SIZE(wid, hgt)    ((3*(wid)*(hgt)) >> 1)
#define FRAME_422_SIZE(wid, hgt)    ((  (wid)*(hgt)) << 1)
#define FRAME_444_SIZE(wid, hgt)     (3*(wid)*(hgt))

#define PIC_DISPLAYED      0x80000000
#define PIC_PREV_GROUP     0x40000000

#define MB_MOTION_FWD_BIT  0x80
#define MB_MOTION_BCK_BIT  0x40
#define MB_PATTERN_BIT     0x20
#define MB_INTRA_BIT       0x10
#define MB_QUANT_BIT       0x08
#define MB_DCT_TYPE_BIT    0x04
#define MB_MISMATCH_BIT    0x01

#define MB_MOTION_FWD(ptr_to_SLICE) \
   (ptr_to_SLICE->mb_flags & 0x80)
#define MB_MOTION_BCK(ptr_to_SLICE) \
   (ptr_to_SLICE->mb_flags & 0x40)
#define MB_PATTERN(ptr_to_SLICE) \
   (ptr_to_SLICE->mb_flags & 0x20)
#define MB_INTRA(ptr_to_SLICE) \
   (ptr_to_SLICE->mb_flags & 0x10)
#define MB_QUANT(ptr_to_SLICE) \
   (ptr_to_SLICE->mb_flags & 0x08)
#define MB_DCT_TYPE(ptr_to_SLICE) \
   (ptr_to_SLICE->mb_flags & 0x04)

#define PRED_TYPE_FIELD_BASED    0
#define PRED_TYPE_FRAME_BASED    1
#define PRED_TYPE_DUAL_PRIME     2
#define PRED_TYPE_MC_16x8        3

#define MV_FORMAT_FIELD(ptr_to_SLICE) \
   (ptr_to_SLICE->pred_type != 1)
#define MV_FORMAT_FRAME(ptr_to_SLICE) \
   (ptr_to_SLICE->pred_type == 1)


#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   s16 QFS[64],            /* DCT coefficients / block values */
       vector[2][2][2],    /* motion vectors           (r, s, t) */
       PMV[2][2][2],       /* motion vector predictors (r, s, t) */
       dct_pred[3],        /* intra DC coefficient predictors (cc) */
       signed_level;       /* stored DCT coefficient */

   u16 mb_wid, mb_hgt,     /* size of reconstructed picture in MB units */
       mb_row, mb_col,     /* location of MB in the bitstream */
       cur_row, cur_col;   /* location of MB in the frame buffer */
       
   SEQ_HEADER *seq_hdr;
   PIC_HEADER *pic_hdr;

   CLK_VAL offset, cur_pts, 
           pes_pts, ref_ptsF, ref_ptsB;
       
   s32 frame_delay,
       move[4];
   u32 ref_tempF,
       ref_tempB,
       cur_temp,
       frame_wid,
       luma_size,
       frame_size;
   u8 *ref_buf,            /* buffer holding the 2 reference frames */
      *ref_frameF,         /* forward prediction reference frame */
      *ref_blkF,           /* current block in fwd ref frame */
      *ref_frameB,         /* backward prediction reference frame */
      *ref_blkB,           /* current block in bck ref frame */
      *pred_frame,         /* this is frame being created */
      *pred_blk,           /* current block in pred frame */
      *mv_blk[4];
      
   u8  bit_shift, data, q_scale_code,
       mb_flags, pred_type, mv_count,
       r, s, t, i, cc, dct_size,
       last_frame, fields_done,
       half_pixel[4], buf[2];
} SLICE;

/* Processes a slice in data_stream.
      INPUT: data_stream - Points to the data stream holding the slice.
                             This stream does not have to contain the
                             entire slice. Repeated calls to this
                             function can be used to read the slice
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: seq_hdr - Points to the SEQ_HEADER of the mpeg2 stream.
              pic_hdr - Points to the PIC_HEADER of the slice.
              cur_pos - Used by MPEG2_VIDEO_MODULE.
              code    - Used by MPEG2_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 SLICE_read(const u8    *data_stream,
                     u32    data_len,
                     SLICE *slice,
                     u32   *cur_pos,
                     u16   *code);

#ifdef __cplusplus
};
#endif

#endif
