/***************************************************************************
                        MPEG_audio.h  -  description
                             -------------------
   Functions for handling mpeg audio frames.
                             -------------------
    begin                : Thu Apr 7 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _MPEG_AUDIO_H_
#define _MPEG_AUDIO_H_

#include "common/typedefs.h"
#include "MPEG2_SYS.h"

#define MPEG_AUDIO_SYNC_BYTE     0xFF
#define MPEG_AUDIO_LAYER1_BYTE   0xFF
#define MPEG_AUDIO_LAYER2_BYTE   0xFD
#define MPEG_AUDIO_LAYER3_BYTE   0xFB

#define MPEG_AUDIO_STATE_MASK             0xF000
#define MPEG_AUDIO_STATE_INC              0x1000
#define MPEG_AUDIO_next_frame_STATE       0x0000
#define MPEG_AUDIO_next_frame_END_STATE \
   (MPEG_AUDIO_next_frame_STATE+MPEG_AUDIO_STATE_INC)

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8  layer,
/*
      4   syncword (end)  <'1111'>
      1   ID   <0> reserved
               <1> MPEG audio
      2   layer   <'00'> reserved
                  <'01'> LAYER_3
                  <'10'> LAYER_2
                  <'11'> LAYER_1
      1   no_protection                Set when the crc data is not included 
                                         in the bitstream.
*/
       num_ch,             /* Number of audio channels in the frame */
       sblimit,            /* Number of sub-bands in the frame */
       bound;              /* Sub-band where intensity stereo begins */

   u32 sample_rate,        /* in Hz */
       frame_len;          /* Total number of bytes in this frame */   
   u16 state,
       payload_len16;      /* Number of bytes in the payload */
   const
   u8 *payload;
   
   CLK_VAL  offset, pes_pts,
           *a_sync;
} MPEG_AUDIO_FRAME;


/* Find the first MPEG audio frame in the given data stream.
      INPUT: data_stream - Points to the data to look through.
             data_len    - Number of bytes in data_stream.
             frame       - Pointer to an MPEG_AUDIO_FRAME.
      RETURN: Success         - Returns number of bytes read from
                                  data_stream.
              Frame not found - 0 */
u32 MPEG_AUDIO_next_frame(const u8         *data_stream,
                          u32               data_len,
                          MPEG_AUDIO_FRAME *frame);

#ifdef __cplusplus
};
#endif

#endif

