/***************************************************************************
                           sound.h  -  description
                             -------------------
   Sound module function prototypes.
                             -------------------
    begin                : Wed May 11 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __SOUND_H__
#define __SOUND_H__

#include "common/typedefs.h"

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif


#ifndef _NO_51_SURROUND_
# define SUR51_ORDER_FLFT_FCEN_FRGT_RLFT_RRGT_LFE  0
# define SUR51_ORDER_FLFT_FRGT_FCEN_LFE_RLFT_RRGT  1
# define SUR51_ORDER_FLFT_FRGT_RLFT_RRGT_FCEN_LFE  2

# define SUR51_CH_FLFT  0
# define SUR51_CH_FCEN  1
# define SUR51_CH_FRGT  2
# define SUR51_CH_RLFT  3
# define SUR51_CH_RRGT  4
# define SUR51_CH_LFE   5
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Prepares the sound device for the incoming audio stream.
      INPUT: num_ch      - Number of audio channels.
             sample_rate - Sample rate (Hz)
      RETURN: Success - Number of audio channels.
              Failure - 0 */
u32  SOUND_init(u32 num_ch, u32 sample_rate);

/* Returns the buffer that will hold the audio samples. Must
     call SOUND_init() function before calling this one.
      INPUT: samples - Number of samples the buffer will hold.
      RETURN:  Success - Pointer to the buffer.
               Failure - NULL */
s16 *SOUND_buf(u32 samples);

/* Called when the sound is no longer needed. */
void SOUND_done();

/* Called when the samples in the buffer should be played. */
void SOUND_play();

/* Returns the number of samples left in the buffer. */
u32  SOUND_samples_left();

/* This function is used to explicitly stop the sound from 
    being heard.  Continuously running systems (ie. DirectSound)
    need this to prevent users from hearing annoying looping 
    sounds when there are no new samples being sent to the 
    sound card (during a pause). 
      INPUT: wait - Set when you want the sound to start after
                     this function is called again with wait=0.
                    Unset when you want the sound to start at
                     the next SOUND_play() call. */
void SOUND_stop(int wait);

#ifndef _NO_51_SURROUND_
/* Sets the order of the 5.1 surround channels to one of the 
    SUR51_ORDER_* values. */
void SOUND_51_SURROUND_set_order(u32 order);

/* Returns the position of the requested channel. */
u32  SOUND_51_SURROUND_get_pos(u32 channel);
#endif

#ifdef __cplusplus
};
#endif

#endif

 
