/***************************************************************************
                         PIC_LAYER.h  -  description
                             -------------------
   Stuff needed for handling the VC1 picture layer.
                             -------------------
    begin                : Tue May 23 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PIC_LAYER_H__
#define __PIC_LAYER_H__

#include "common/typedefs.h"
#include "SEQ_LAYER.h"
#include "bitplane.h"
#include "inv_trans.h"


#define FRAME_420_SIZE(wid, hgt)    ((3*(wid)*(hgt)) >> 1)


#define VC1_PIC_TYPEp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->type & 0xF0)
#define VC1_PIC_TYPEs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.type  & 0xF0)
#define VC1_SKIP_FRAME  0x00
#define VC1_I_FRAME     0x10
#define VC1_P_FRAME     0x20
#define VC1_B_FRAME     0x40
#define VC1_BI_FRAME    0x50

#define VC1_PIC_RANGE_REDp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->type & 0x08)
#define VC1_PIC_RANGE_REDs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.type  & 0x08)

#define VC1_PIC_UNIFORM_QUANTp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->type & 0x04)
#define VC1_PIC_UNIFORM_QUANTs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.type  & 0x04)

#define VC1_AC_MODE3_TABp(ptr_to_PIC_LAYER) \
   ((ptr_to_PIC_LAYER->type >> 1) & 0x01)
#define VC1_AC_MODE3_TABs(struct_PIC_LAYER) \
   ((struct_PIC_LAYER.type  >> 1) & 0x01)

#define VC1_PIC_LOW_MOTIONp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->type & 0x01)
#define VC1_PIC_LOW_MOTIONs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.type  & 0x01)


#define VC1_AC_MODE3p(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->trans_tab >= 0x80)
#define VC1_AC_MODE3s(struct_PIC_LAYER) \
   (struct_PIC_LAYER.trans_tab  >= 0x80)
#define VC1_AC_MODE3       0x80

#define VC1_AC_MODE3_LASTp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->trans_tab & 0x40)
#define VC1_AC_MODE3_LASTs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.trans_tab  & 0x40)
#define VC1_AC_MODE3_LAST  0x40

#define VC1_AC_MODE2p(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->trans_tab & 0x20)
#define VC1_AC_MODE2s(struct_PIC_LAYER) \
   (struct_PIC_LAYER.trans_tab  & 0x20)
#define VC1_AC_MODE2       0x20

#define VC1_PIC_TRANS_AC1p(ptr_to_PIC_LAYER) \
   ((ptr_to_PIC_LAYER->trans_tab >> 3) & 0x03)
#define VC1_PIC_TRANS_AC1s(struct_PIC_LAYER) \
   ((struct_PIC_LAYER.trans_tab  >> 3) & 0x03)

#define VC1_PIC_TRANS_AC2p(ptr_to_PIC_LAYER) \
   ((ptr_to_PIC_LAYER->trans_tab >> 1) & 0x03)
#define VC1_PIC_TRANS_AC2s(struct_PIC_LAYER) \
   ((struct_PIC_LAYER.trans_tab  >> 1) & 0x03)

#define VC1_PIC_TRANS_DCp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->trans_tab & 0x01)
#define VC1_PIC_TRANS_DCs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.trans_tab  & 0x01)

#define VC1_PIC_MV_TABLEp(ptr_to_PIC_LAYER) \
   ((ptr_to_PIC_LAYER->mv_cbp_tab >> 6) & 0x03)
#define VC1_PIC_MV_TABLEs(struct_PIC_LAYER) \
   ((struct_PIC_LAYER.mv_cbp_tab  >> 6) & 0x03)

#define VC1_PIC_CBP_TABLEp(ptr_to_PIC_LAYER) \
   ((ptr_to_PIC_LAYER->mv_cbp_tab >> 4) & 0x03)
#define VC1_PIC_CBP_TABLEs(struct_PIC_LAYER) \
   ((struct_PIC_LAYER.mv_cbp_tab  >> 4) & 0x03)

#define VC1_AC_PREDp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->mv_cbp_tab & 0x03)
#define VC1_AC_PREDs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.mv_cbp_tab  & 0x03)
#define VC1_AC_PRED_NONE   0x00
#define VC1_AC_PRED_LEFT   0x01
#define VC1_AC_PRED_TOP    0x03


#define VC1_PIC_DQUANT_FRMp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->dquant >= 0x80)
#define VC1_PIC_DQUANT_FRMs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.dquant  >= 0x80)

#define VC1_PIC_DQUANT_PROFILEp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->dquant & 0x60)
#define VC1_PIC_DQUANT_PROFILEs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.dquant  & 0x60)
#define VC1_DQUANT_PROFILE_4EDGES   0x00
#define VC1_DQUANT_PROFILE_2EDGES   0x20
#define VC1_DQUANT_PROFILE_1EDGE    0x40
#define VC1_DQUANT_PROFILE_ALL_MBS  0x60

#define VC1_PIC_DQUANT_EDGEp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->dquant & 0x18)
#define VC1_PIC_DQUANT_EDGEs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.dquant  & 0x18)
#define VC1_DQUANT_EDGE_LEFT        0x00
#define VC1_DQUANT_EDGE_TOP         0x08
#define VC1_DQUANT_EDGE_RIGHT       0x10
#define VC1_DQUANT_EDGE_BOTTOM      0x18

#define VC1_PIC_DQUANT_BILEVELp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->dquant & 0x10)
#define VC1_PIC_DQUANT_BILEVELs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.dquant  & 0x10)

#define VC1_PIC_DQUANT_INFRAMEp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->dquant & 0x04)
#define VC1_PIC_DQUANT_INFRAMEs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.dquant  & 0x04)


#define VC1_VST_TYPEp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->VST & 0x03)
#define VC1_VST_TYPEs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.VST  & 0x03)
#define VC1_VST_TYPE_8x8   0x00
#define VC1_VST_TYPE_8x4   0x01
#define VC1_VST_TYPE_4x8   0x02
#define VC1_VST_TYPE_4x4   0x03

#define VC1_VST_LEVELp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->VST & 0x0C)
#define VC1_VST_LEVELs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.VST  & 0x0C)
#define VC1_VST_LEVEL_NONE    0x00
#define VC1_VST_LEVEL_FRAME   0x04
#define VC1_VST_LEVEL_MACRO   0x0C
#define VC1_VST_LEVEL_BLOCK   0x08

#define VC1_VST_CODEDp(ptr_to_PIC_LAYER) \
   (ptr_to_PIC_LAYER->VST & 0xF0)
#define VC1_VST_CODEDs(struct_PIC_LAYER) \
   (struct_PIC_LAYER.VST  & 0xF0)
#define VC1_VST_CODED_8x8_ALL    0x80
#define VC1_VST_CODED_8x4_TOP    0x80
#define VC1_VST_CODED_8x4_BOT    0x40
#define VC1_VST_CODED_8x4_ALL    0xC0
#define VC1_VST_CODED_4x8_LFT    0x80
#define VC1_VST_CODED_4x8_RGT    0x40
#define VC1_VST_CODED_4x8_ALL    0xC0
#define VC1_VST_CODED_4x4_TOPLFT 0x80
#define VC1_VST_CODED_4x4_TOPRGT 0x40
#define VC1_VST_CODED_4x4_BOTLFT 0x20
#define VC1_VST_CODED_4x4_BOTRGT 0x10
#define VC1_VST_CODED_4x4_ALL    0xF0

#define VC1_VST_CODED_BLK_INIT   0x10
#define VC1_VST_CODED_SUB_INIT   0x20


#define VC1_MVMODE_1MV              0x01
#define VC1_MVMODE_1MV_HALFPIX      0x03
#define VC1_MVMODE_1MV_HALFPIX_BI   0x07
#define VC1_MVMODE_MIXED_MV         0x08
#define VC1_MVMODE_INT_COMP         0x10
#define VC1_MVMODE_CODED            0x40
#define VC1_MVMODE_INTRA            0x80

#define VC1_MVMODE_HALFPIX          0x02
#define VC1_MVMODE_BILINEAR         0x04

#define VC1_BP_RAW_MVTYPE           0x80
#define VC1_BP_RAW_MBDIRECT         0x40
#define VC1_BP_RAW_MBSKIP           0x20


#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8   blk[(11*11) + 7];
   s16  coeff[64],
        DCpred_Y[   (VC1_MAX_FRAME_WID/4)  +  2 + 2],
        DCpred_C[   (VC1_MAX_FRAME_WID/8)  +  2],
        ACpred_Y[(7*(VC1_MAX_FRAME_WID/8)) + 14],
        ACpred_C[(7*(VC1_MAX_FRAME_WID/8)) + 14],
        MV[         (VC1_MAX_FRAME_WID/2)  +  2]
#ifdef USE_PRECISE_OVERLAP
      , Ocoeff[     (VC1_MAX_FRAME_WID*4)  +192]
#endif
        ;
   s32  MVinc[6];
   u16  MVout_qp[6];
   u8   MVtype[VC1_BITPLANE_SIZE],
        MBdirect[VC1_BITPLANE_SIZE],
        MBskip[VC1_BITPLANE_SIZE],
        CBPbuf[VC1_MAX_FRAME_WID/16],
        LUT[256*2];
   const
   u8  *tab, *ACrunlevel, *zz;

   u32 /* ref_tempF,
       ref_tempB,
       cur_temp, */
       frame_wid,
       luma_size,
       frame_size;
   u8 /*frame,              * non-reference (B) frame */
      *ref_frameF,         /* forward prediction reference frame */
      *ref_blkF,           /* current block in fwd ref frame */
      *ref_frameB,         /* backward prediction reference frame */
      *ref_blkB,           /* current block in bck ref frame */
      *pred_frame,         /* this is frame being created */
      *pred_blk;           /* current block in pred frame */

   u16  BPi, x, y, b;
   s16  DCpred_def, c;
   u8   last_frame,
        type,
/*
      4   PType                        Picture type.
      1   RangeRed_Frm                 Set when range reduction is used for
                                        this frame.
      1   PQuantizer                   Set when the uniform quantizer is used.
                                        Non-uniform quantizer is used otherwise.
      1   ACmode3_Tab
      1   LowMotion
*/
        trans_tab,
/*
      1   ACmode3
      1   ACmode3_last
      1   ACmode2
      2   TransAC_Frm                  AC transform code table index.
      2   TransAC_Frm2                 Secondary AC transform code table index.
      1   TransDC_Tab                  DC transform code table index.
*/
        mv_cbp_tab,
/*
      2   MV_Tab
      2   CBP_Tab
      2   unused
      2   ACpred
*/
        dquant,
/*
      1   DQuantFrm                    When not set, this picture shall only
                                        be quantized with PQuant.
      2   DQprofile                    Specifies where quantization step
                                        changes are allowed to occur, within
                                        this picture.
                      <0> 'All 4 Edges'
                      <1> 'Double Edge'
                      <2> 'Single Edge'
                      <3> 'All Macroblocks'
      2   DQedge                       Specifies the edge that is quantized by
                                        AltPQuant.
                   <0> left
                   <1> top
                   <2> right
                   <3> bottom
      1   DQbilevel
      1   DQuant_InFrame
*/
        VST,
/*
      4   Coded
      2   Level
      2   Type
*/
        BPraw, BPbit,
        Bfraction,
        Pquant, Pquant_alt, Mquant, DCstep, DCstep_org,
        MVmode, MVsize, MV4mode,
        Yscale, Yshift, overlap,
        CBP, CBP_l, CBP_lt, CBP_t, rnd,
        ACidx, level_size, run_size;
} PIC_LAYER;


/* Read the first part of the PIC_LAYER data.
      INPUT: seq - Pointer to a SEQ_LAYER structure.
             pic - Pointer to a PIC_LAYER structure. */
void VC1_VIDEO_read_PIC_LAYER1(SEQ_LAYER *seq, PIC_LAYER *pic);

/* Read the second part of the PIC_LAYER data.
      INPUT: seq - Pointer to a SEQ_LAYER structure.
             pic - Pointer to a PIC_LAYER structure. */
void VC1_VIDEO_read_PIC_LAYER2(SEQ_LAYER *seq, PIC_LAYER *pic);

extern u32 VC1_frame_cnt;
#define VC1_TRACE_BEG_BLK  if (VC1_frame_cnt >= 0x80000000) {
#define VC1_TRACE_END_BLK  }

#ifdef __cplusplus
};
#endif

#endif
