/***************************************************************************
                       file_actions.c  -  description
                             -------------------
    begin                : Sat Feb 5 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>

#include "winmain.h"
#include "resource.h"
#include "FileDLG.h"
#include "ListDLG.h"
#include "file_actions.h"
#include "control_actions.h"
#include "StatsDLG.h"
#include "MPEG2_SYS.h"
#include "mpeg_es.h"
#include "mpeg_ps.h"
#include "mpeg_ts.h"
#include "commands.h"
#include "detect_mpeg.h"
#include "PMT_SECTION.h"
#include "common/Thread.h"
#include "common/File.h"

static THREAD playing = NULL;
static void * THREAD_FUNC_TYPE start_playing(void *arg);

static void start_ES(FILE_PTR fp, u8 stream_type);
static void start_TS(FILE_PTR fp);
static void start_PS(FILE_PTR fp);
static void init_command();
static void alter_control_state(UINT state, int optional);


void FILE_OPEN_ACTION()
{
   FILE_CLOSE_ACTION();

   FileDLG.lpstrTitle  = "Open MPEG File(s)";
   FileDLG.lpstrFilter = "MPEG Files\0"
                            "*.mpg;*.mpeg;*.m2t;*.vob;"
                            "*.m2v;*.m1v;*.mp2;*.ac3\0"
                         "WMV Files\0"
                            "*.wmv\0"
                         "All Files (*.*)\0*.*\0"
                            "\0";
   FileDLG.Flags      &= ~FILEDLG_SAVEFLAGS;
   FileDLG.Flags      |=  FILEDLG_LOADFLAGS;
   FileDLG_filename[0] = '\0';
   if (GetOpenFileName(&FileDLG)) {
      sys.data.command = CMD_PLAY;
      playing = new_THREAD(start_playing, NULL);
   }
}

void FILE_CLOSE_ACTION()
{
   if (playing) {
      HMENU menu = GetMenu(win);

      sys.data.command = CMD_STOP;
      broadcast_CONDVAR(sys.cv);

      join_THREAD(playing, NULL);
      del_THREAD(playing);
      playing = NULL;
   }
}

UINT FILE_LOOP_ACTION(int invert)
{
   MENUITEMINFO info;
   HMENU        menu = GetMenu(win);

   info.cbSize = sizeof(MENUITEMINFO);
   info.fMask  = MIIM_STATE;
   if (GetMenuItemInfo(menu, ID_MAIN_MENU_FILE_LOOP, FALSE, &info)) {
      if (invert) {
         info.fState ^= MFS_CHECKED;
         SetMenuItemInfo(menu, ID_MAIN_MENU_FILE_LOOP, FALSE, &info);
      }
      return (info.fState & MFS_CHECKED);
   }
   return 0;
}


#define MAX_NAMELEN 256
static void * THREAD_FUNC_TYPE start_playing(void *arg)
{
   FILE_BYTE_COUNT len;
   FILE_PTR        fp;
   char            filename[MAX_NAMELEN];
   const char     *nxt,
                  *cur      = FileDLG_filename,
                  *err_msg  = NULL;
   size_t          path_len = strlen(FileDLG_filename);

PLAY_LOOP:
   while ( (sys.data.command != CMD_STOP)&&
           (nxt = strchr(cur, '\0'))&&(nxt > cur) ) {
      if ( (cur > FileDLG_filename)&&(*(nxt+1) == '\0') )
         break;

      cur = nxt+1;
      len = (FILE_BYTE_COUNT)strlen(cur);
      if ( (len > 0)&&((path_len+(size_t)len) > (MAX_NAMELEN-2)) )
         err_msg = "File name is too long.";
      else {
         if (len > 0) {
            sprintf(filename, "%s\\%s", FileDLG_filename, cur);
            FILE_OPEN_READ(fp, filename);
         }
         else {
            FILE_OPEN_READ(fp, FileDLG_filename);
         }

         if (FILE_PTR_BAD(fp))
            err_msg = "Could not open the file.";
         else {
            FILE_READ(fp, sys.buf, BUFFER_SIZE, len);

            if (len < (5*TS_PKT_SIZE))
               err_msg = "Could not read enough data from the file.";
            else {
               switch (detect_mpeg(sys.buf, (u32)len)) {
               case MPEG_TYPE_ES_V1:
                  start_ES(fp, STREAM_TYPE_MPEG1_VIDEO);
                  break;
               case MPEG_TYPE_ES_A1:
                  start_ES(fp, STREAM_TYPE_MPEG1_AUDIO);
                  break;
               case MPEG_TYPE_ES_V2:
                  start_ES(fp, STREAM_TYPE_MPEG2_VIDEO);
                  break;
               case MPEG_TYPE_ES_A2:
                  start_ES(fp, STREAM_TYPE_MPEG2_AUDIO);
                  break;
               case MPEG_TYPE_ES_AC3:
                  start_ES(fp, STREAM_TYPE_ATSC_AC3_AUDIO);
                  break;

               case MPEG_TYPE_TS:
                  start_TS(fp);
                  break;

               case MPEG_TYPE_PS:
                  start_PS(fp);
                  break;

               case MPEG_TYPE_ES_WMV:
                  start_ES(fp, STREAM_TYPE_WMV_DATA);
                  break;

               case MPEG_TYPE_UNKNOWN:
               default:
                  err_msg = "Could not determine the file type.";
               }
            }
         }
      }
      MPEG2_SYS_reset_PIDs();

      if (FILE_PTR_OK(fp))
         FILE_CLOSE(fp);
      if (err_msg) {
         MessageBox(win, err_msg,
                    ERRMSG_TITLE, (MB_OK|MB_ICONEXCLAMATION));
         err_msg = NULL;
      }
   }

   if ( (FILE_LOOP_ACTION(0))&&(sys.data.command != CMD_STOP) ) {
      cur = FileDLG_filename;
      goto PLAY_LOOP;
   }

   sys.data.command = CMD_STOP;
   MPEG2_SYS_modules_done();
   InvalidateRect(win, NULL, TRUE);

   exit_THREAD(NULL);
   return NULL;
}
#undef MAX_NAMELEN


static void start_ES(FILE_PTR fp, u8 stream_type)
{
   if (sys.data.command == CMD_STOP)
      return;

   FILE_REWIND(fp);
   alter_control_state(MF_ENABLED, 0);
   init_command();

   MPEG_ES_Play(fp, stream_type);
#ifdef _RECORD_STATS_
   if ( (sys.data.command != CMD_STOP)&&
        (win_mode == WIN_MODE_NORMAL) )
      StatsDLG();
#endif

   alter_control_state(MF_GRAYED, 0);
}

static void start_TS(FILE_PTR fp)
{
   /* find something to play */
   FILE_REWIND(fp);
   MPEG_TS_Find_Program(fp);
   if (!sys.program_number)
      MPEG_TS_Create_Program(fp);
   if (sys.program_number) {
      FILE_REWIND(fp);
      alter_control_state(MF_ENABLED, 2);
RESTART_TS:
      init_command();

      MPEG2_SYS_Play(fp, MPEG_TS_Play);

      if (win_mode >= WIN_MODE_SEL_PROG) {
         sys.data.command = CMD_PLAY;
         InvalidateRect(win, NULL, FALSE);
         switch (win_mode) {
         case WIN_MODE_SEL_PROG:
            ListDLG(fp, 1, 1);
            break;
         case WIN_MODE_SEL_STR:
            ListDLG(fp, 0, 1);
         }
         win_mode = WIN_MODE_NORMAL;
         goto RESTART_TS;
      }
#ifdef _RECORD_STATS_
      if ( (sys.data.command != CMD_STOP)&&
           (win_mode == WIN_MODE_NORMAL) )
         StatsDLG();
#endif

      alter_control_state(MF_GRAYED, 2);
   }
   else {
      MessageBox(win, "Cannot find a TS Program.",
                 ERRMSG_TITLE, (MB_OK|MB_ICONEXCLAMATION));
   }
}

static void start_PS(FILE_PTR fp)
{
   /* find something to play */
   FILE_REWIND(fp);
   MPEG_PS_Create_Program(fp);
   if (sys.program_number) {
      FILE_REWIND(fp);
      alter_control_state(MF_ENABLED, 1);
RESTART_PS:
      init_command();

      MPEG2_SYS_Play(fp, MPEG_PS_Play);

      if (win_mode == WIN_MODE_SEL_STR) {
         sys.data.command = CMD_PLAY;
         InvalidateRect(win, NULL, FALSE);
         ListDLG(fp, 0, 0);
         win_mode = WIN_MODE_NORMAL;
         goto RESTART_PS;
      }
#ifdef _RECORD_STATS_
      if ( (sys.data.command != CMD_STOP)&&
           (win_mode == WIN_MODE_NORMAL) )
         StatsDLG();
#endif

      alter_control_state(MF_GRAYED, 1);
   }
   else {
      MessageBox(win, "Cannot find anything to play.",
                 ERRMSG_TITLE, (MB_OK|MB_ICONEXCLAMATION));
   }
}

static void init_command()
{
   u32 cmd;

   if (sys.data.command == CMD_PAUSE) {
      sys.data.command = CMD_PLAY;
      cmd              = CMD_PAUSE;
   }
   else if (sys.data.command == CMD_SLOW_FWD) {
      sys.data.command = CMD_PLAY;
      cmd              = CMD_SLOW_FWD;
   }
   else {
      sys.data.command = CMD_PAUSE;
      cmd              = CMD_PLAY;
   }
   sys.jump_offset = 0;
   CONTROL_PLAY_ACTION(MFS_ENABLED);
   sys.data.command = cmd;
}

static void alter_control_state(UINT state, int optional)
{
   HMENU menu = GetMenu(win);

   EnableMenuItem(menu, ID_MAIN_MENU_FILE_CLOSE, 
                  (MF_BYCOMMAND | state));
   EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_PLAY,
                  (MF_BYCOMMAND | state));
   EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_SLOW,
                  (MF_BYCOMMAND | state));
   EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_JUMP_FWD, 
                  (MF_BYCOMMAND | state));
   EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_JUMP_BCK, 
                  (MF_BYCOMMAND | state));
   EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_JUMP2_FWD, 
                  (MF_BYCOMMAND | state));
   EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_JUMP2_BCK, 
                  (MF_BYCOMMAND | state));
   EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_JUMP2_START, 
                  (MF_BYCOMMAND | state));

   if (optional >= 1) {
      EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_SELECTSTREAMS, 
                     (MF_BYCOMMAND | state));
   }
   if (optional >= 2) {
      EnableMenuItem(menu, ID_MAIN_MENU_CONTROL_SELECTPROGRAM, 
                     (MF_BYCOMMAND | state));
   }
}

