/***************************************************************************
                         x11_main.c  -  description
                             -------------------
   Handle the all the x11 stuff.
                             -------------------
    begin                : Fri Mar 4 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef _NO_VIDEO_
# include <X11/keysym.h>
#endif

#ifndef _NO_TERM_INPUT_
# include <unistd.h>
# include <termios.h>
# include <signal.h>

# ifndef TCSASOFT
#  define TCSASOFT  0
# endif
#endif

#include "common/Thread.h"
#include "common/File.h"
#ifdef __DVD_INPUT__
# include "dvd/dvd_input.h"
#endif
#include "detect_mpeg.h"
#include "commands.h"
#include "MPEG2_SYS.h"
#include "PMT_SECTION.h"
#include "x11_main.h"
#include "cmdline_args.h"
#include "file_actions.h"


#ifndef _NO_VIDEO_
Display *X_comm     = NULL;
Window   win;
GC       gc         = None;
int      win_x      = 0,
         win_y      = 0;
u32      win_wid    = 0,
         win_hgt    = 0,
         screen_wid = 0,
         screen_hgt = 0;

static const char *win_title  = "MPEG Audio/Video Player";
static THREAD      win_events = NULL;
static long        events     = NoEventMask;
static int         fs_x       = 0,
                   fs_y       = 0;
static u32         fs_wid     = 0,
                   fs_hgt     = 0;

static void * THREAD_FUNC_TYPE Xevent_loop(void *args);
#endif

#ifndef _NO_TERM_INPUT_
static THREAD      term_events = NULL;

static void * THREAD_FUNC_TYPE TERMevent_loop(void *args);
static void   usr1_sig(int sig);
#endif

int main(int argc, char *argv[])
{
   size_t      len;
   int         idx      = 1;
#ifndef _NO_TERM_INPUT_
   int         run_term = 1;
#endif
   APP_PARAMS  params   = {PLAY_FIRST, 0, 0, 0, 0, 0, 0.0, NULL,
#ifdef __DVD_INPUT__
                           0,
                           DVD_PTR_BAD_VALUE,
#endif
                           FILE_PTR_BAD_VALUE};
   const char *err_msg  = NULL;

   /* allocate memory */
   if (!MPEG2_SYS_alloc(&sys))
      return -1;
#ifndef _NO_VIDEO_
   /* make sure we can use multiple threads with Xlib */
   if (!XInitThreads())
      return -2;
#endif
#ifndef _NO_TERM_INPUT_
   /* use SIGUSR1 to notify the term_events thread of shutdown */
   signal(SIGUSR1, usr1_sig);
   siginterrupt(SIGUSR1, 1);
#endif

   while ((idx = CMDLINE_parse(idx, argc, argv, &params)) > CMDLINE_DONE) {
#ifndef _NO_TERM_INPUT_
      if (!term_events) {
         /* start the event handler for the terminal */
         term_events = new_THREAD(TERMevent_loop, &run_term);
      }
#endif
      printf("%s:\n", params.filename);
      
      sys.jump_offset   = (CLK_VAL)(params.start_time*MPEG2_SYS_CLK_FREQ);
      params.start_time = 0.0;

#ifdef __DVD_INPUT__
      if (DVD_PTR_OK(params.dvd))
         goto MPEG_PS_DETECTED;
#endif
      FILE_READ(params.fp, sys.buf, (TS_PKT_SIZE*MPEG2_SYS_BUF_PKTS), len);
      if (len < (4*TS_PKT_SIZE))
         err_msg = "Could not read enough data from the file.";
      else {
         /* find out what is in the file, and do something with it */
         switch (detect_mpeg(sys.buf, (u32)len)) {
         case MPEG_TYPE_ES_V1:
            printf("MPEG1 Video\n");
            start_ES(params.fp, STREAM_TYPE_MPEG1_VIDEO);
            break;
         case MPEG_TYPE_ES_A1:
            printf("MPEG1 Audio\n");
            start_ES(params.fp, STREAM_TYPE_MPEG1_AUDIO);
            break;
         case MPEG_TYPE_ES_V2:
            printf("MPEG2 Video\n");
            start_ES(params.fp, STREAM_TYPE_MPEG2_VIDEO);
            break;
         case MPEG_TYPE_ES_A2:
            err_msg = "MPEG2 Audio stream.";
            break;
         case MPEG_TYPE_TS:
            printf("MPEG Transport Stream\n");
            start_TS(&params);
            break;
         case MPEG_TYPE_PS:
#ifdef __DVD_INPUT__
MPEG_PS_DETECTED:
#endif
            printf("MPEG Program Stream\n");
            start_PS(&params);
            break;
         case MPEG_TYPE_ES_AC3:
            printf("AC3 Audio Stream\n");
            start_ES(params.fp, STREAM_TYPE_ATSC_AC3_AUDIO);
            break;
         case MPEG_TYPE_ES_WMV:
            printf("WMV Data\n");
            start_ES(params.fp, STREAM_TYPE_WMV_DATA);
            break;
         case MPEG_TYPE_UNKNOWN:
         default:
            err_msg = "Could not determine the file type.";
         }
      }
      MPEG2_SYS_reset_PIDs();

      if (FILE_PTR_OK(params.fp)) {
         FILE_CLOSE(params.fp);
         params.fp = FILE_PTR_BAD_VALUE;
      }
#ifdef __DVD_INPUT__
      else if (DVD_PTR_OK(params.dvd)) {
         DVD_CLOSE(params.dvd);
         params.dvd = DVD_PTR_BAD_VALUE;
         if ( (++dvd_title <= dvd_title_last)&&
              (params.req != DISPLAY_DVD_INFO)&&
              (sys.data.command != CMD_STOP) ) {
            DVD_OPEN_READ(params.dvd, params.filename);
            if (DVD_PTR_OK(params.dvd))
               goto MPEG_PS_DETECTED;
         }
         else {
            DVD_RESET_TITLE_POS(params.dvd);
         }
      }
#endif

      if (err_msg)
         printf("ERROR: %s\n\n", err_msg);
      else if (sys.data.command == CMD_STOP)
         idx = argc;
      else if ( (params.loop)&&(params.req <= PLAY_STREAMS)&&(idx == argc) )
         idx = 1;
   }
   sys.data.command = CMD_STOP;
   MPEG2_SYS_modules_done();
   
#ifndef _NO_TERM_INPUT_
   if (term_events) {
      run_term = 0;              /* this is the shutdown request */
      kill(getpid(), SIGUSR1);   /* interrupt thread, if blocking */

      /* wait for event handler to stop */
      join_THREAD(term_events, NULL);
      del_THREAD(term_events);
   }
#endif

   /* free memory */
   MPEG2_SYS_free(&sys);

   if (idx == CMDLINE_USAGE)
      CMDLINE_usage(argv[0]);
   return idx;
}

#ifndef _NO_VIDEO_
int init_X11(u32 disp_wid, u32 disp_hgt)
{
   if (!X_comm) {
      unsigned long black;
      XEvent        e;

      /* start communications with X server */
      X_comm = XOpenDisplay(NULL);
      if (!X_comm)
         return 0;   /* could not start communications with X server */

      /* make sure window does not get too big */
      screen_wid = (u32)DisplayWidth(X_comm, DefaultScreen(X_comm));
      screen_hgt = (u32)DisplayHeight(X_comm, DefaultScreen(X_comm));
      win_wid    = disp_wid;
      win_hgt    = disp_hgt;
      if ( (disp_wid > screen_wid)||(disp_hgt > screen_hgt) )
         get_win_size(disp_wid, disp_hgt);

      /* create the X11 window */
      black   = BlackPixel(X_comm, DefaultScreen(X_comm));
      win     = XCreateSimpleWindow(X_comm, DefaultRootWindow(X_comm),
                                    0, 0, win_wid, win_hgt,
                                    0, black, black);
      XStoreName(X_comm, win, win_title);
      XSetIconName(X_comm, win, win_title);

      /* put the window on the screen */
      events = StructureNotifyMask;
      XSelectInput(X_comm, win, events);
      XMapWindow(X_comm, win);
      do {
         XNextEvent(X_comm, &e);
      } while ( (e.type != MapNotify)||(e.xmap.window != win) );

      /* start the event handler for the window */
      events |= KeyPressMask | KeyReleaseMask;
      XSelectInput(X_comm, win, events);
      win_events = new_THREAD(Xevent_loop, NULL);
      if (!win_events) {
         XDestroyWindow(X_comm, win);
         XCloseDisplay(X_comm);
         X_comm = NULL;
         return 0;   /* could not start the event handler */
      }
      gc = XCreateGC(X_comm, win, 0, 0);
      XFlush(X_comm);
   }
   else if ( (win_wid != disp_wid)||(win_hgt != disp_hgt) ) {
      win_x  *= 2;        win_y  *= 2;
      win_x  += win_wid;  win_y  += win_hgt;
      win_wid = disp_wid; win_hgt = disp_hgt;

      get_win_size(win_x, win_y);
#if 0
      Window root, child;
      int    x, y;
      /* get current window position */
      XGetGeometry(X_comm, win, &root, &x, &x, &x, &x, &x, &x);
      XTranslateCoordinates(X_comm, win, root, 0, 0, &x, &y, &child);

      /* make sure new window size is completely visible */
      if ((x+(int)win_wid) > (int)screen_wid)
         x = (int)screen_wid - (int)win_wid;
      if ((y+(int)win_hgt) > (int)screen_hgt)
         y = (int)screen_hgt - (int)win_hgt;

      /* resize window */
      XMoveResizeWindow(X_comm, win, x, y, win_wid, win_hgt);
      XFlush(X_comm);
#endif
   }
   return 1;
}

void free_X11()
{
   if (X_comm) {
      /* remove window */
      XFreeGC(X_comm, gc);
      XDestroyWindow(X_comm, win);
      XFlush(X_comm);

      /* wait for event handler to stop */
      join_THREAD(win_events, NULL);
      del_THREAD(win_events);
      win_events = NULL;

      /* end communications with X server */
      XCloseDisplay(X_comm);
      X_comm = NULL;
   }
}

void get_win_size(u32 disp_wid, u32 disp_hgt)
{
   u32 new_wid, new_hgt;

   if (disp_wid > screen_wid)
      disp_wid = screen_wid;
   if (disp_hgt > screen_hgt)
      disp_hgt = screen_hgt;
      
   new_wid  = disp_hgt; new_hgt  = disp_wid;
   new_wid *= win_wid;  new_hgt *= win_hgt;
   new_wid /= win_hgt;  new_hgt /= win_wid;
   
   if (new_wid >= disp_wid) {
      win_wid = disp_wid;
      win_x   = 0;
   }
   else {
      win_wid = new_wid;
      win_x   = (int)(disp_wid - new_wid) / 2;
   }
   
   if (new_hgt >= disp_hgt) {
      win_hgt = disp_hgt;
      win_y   = 0;
   }
   else {
      win_hgt = new_hgt;
      win_y   = (int)(disp_hgt - new_hgt) / 2;
   }
}


static void fullscreen_change()
{
   int x, y;
   u32 w, h;

   if (fs_wid) {  /* restore window size */
      x = fs_x;   y = fs_y;
      w = fs_wid; h = fs_hgt;

      fs_wid = 0;
   }
   else {         /* go fullscreen (not really fullscreen, just make window really big) */
      Window root, child;

      /* get current window position */
      XGetGeometry(X_comm, win, &root, &fs_x, &fs_y, &x, &x, &x, &x);
      XTranslateCoordinates(X_comm, win, root, fs_x, fs_y, &fs_x, &fs_y, &child);
      fs_wid = (win_x*2) + win_wid;
      fs_hgt = (win_y*2) + win_hgt;

      x = -4;         y = -24;
      w = screen_wid; h = screen_hgt;
   }

   /* resize window */
   XMoveResizeWindow(X_comm, win, x, y, w, h);
   XFlush(X_comm);
   get_win_size(w, h);
}
#endif

static void pause_request()
{
   sys.data.command = CMD_PAUSE;
#ifdef __DVD_INPUT__
   printf("\r                                        \rPAUSE @ ");
#else
   printf("\r                          \rPAUSE @ ");
#endif
   display_time( ((sys.data.a_pos > sys.data.v_pos)?
                   sys.data.a_pos : sys.data.v_pos) /
                 (double)MPEG2_SYS_CLK_FREQ);
#ifdef __DVD_INPUT__
   if (dvd_title_last < 0xFFFF)
      printf("  Chapter: %u ", dvd_chapter);
/*   if (dvd_time) {
      printf("  PTS: ");
      display_time((double)dvd_time / (double)MPEG2_SYS_CLK_FREQ);
   }*/
#endif
   fflush(stdout);
}

static void jump_request(CLK_VAL offset)
{
   lock_MUTEX(sys.m, 1);
   while (sys.data.command & CMD_JUMP) {
      wait_CONDVAR(sys.cv, sys.m);
   }
   sys.jump_offset  = offset;
   sys.data.command = CMD_JUMP;
   unlock_MUTEX(sys.m);
}

#ifdef __DVD_INPUT__
static void jump_chapter_request(CLK_VAL offset)
{
   if (dvd_title_last == 0xFFFF)
      return;

   lock_MUTEX(sys.m, 1);
   while (sys.data.command & CMD_JUMP) {
      wait_CONDVAR(sys.cv, sys.m);
   }
   sys.jump_offset  = offset;
   sys.data.command = CMD_JUMP_CHAPTER;
   unlock_MUTEX(sys.m);
}
#endif

#ifndef _NO_VIDEO_
/* Event handler for the X11 window */
static void * THREAD_FUNC_TYPE Xevent_loop(void *args)
{
   XEvent e;
   u32    last_cmd;
   int    process_events = 1,
          ctrl           = 0;

   do {
      XNextEvent(X_comm, &e);
      if (e.xany.window == win) {
         switch (e.type) {
         case ConfigureNotify:
            get_win_size((u32)e.xconfigure.width,
                         (u32)e.xconfigure.height);
            break;
            
         case KeyPress:
            switch (XLookupKeysym((XKeyEvent *)&e, 0)) {
            case XK_Control_L:
            case XK_Control_R:
               ctrl = 1;
            }
            break;
            
         case KeyRelease:
            last_cmd = sys.data.command;
            
            switch (XLookupKeysym((XKeyEvent *)&e, 0)) {
            case XK_Control_L:
            case XK_Control_R:
               ctrl = 0;
               break;
            case XK_space:
            case XK_p:
               if (sys.data.command != CMD_PAUSE)
                  pause_request();
               else
                  sys.data.command = CMD_PLAY;
               break;
            case XK_Left:
            case XK_KP_Left:
               jump_request(-CMD_JUMP_SECS*MPEG2_SYS_CLK_FREQ);
               break;
            case XK_Up:
            case XK_KP_Up:
               jump_request(-CMD_JUMP2_SECS*MPEG2_SYS_CLK_FREQ);
               break;
            case XK_Home:
            case XK_KP_Home:
               jump_request( 0);
               break;
            case XK_Right:
            case XK_KP_Right:
               jump_request( CMD_JUMP_SECS*MPEG2_SYS_CLK_FREQ);
               break;
            case XK_Down:
            case XK_KP_Down:
               jump_request( CMD_JUMP2_SECS*MPEG2_SYS_CLK_FREQ);
               break;
#ifdef __DVD_INPUT__
            case XK_Page_Up:
            case XK_KP_Page_Up:
               jump_chapter_request(-1);
               break;
            case XK_Page_Down:
            case XK_KP_Page_Down:
               jump_chapter_request( 1);
               break;
#endif
            case XK_s:
               if (sys.data.command != CMD_SLOW_FWD) {
                  if (sys.data.v_pos > 0)
                     sys.data.command = CMD_SLOW_FWD;
               }
               else
                  sys.data.command = CMD_PLAY;
               break;
            case XK_f:
               fullscreen_change();
               break;
            case XK_w:
               if (!ctrl)
                  break;
            case XK_Escape:
               sys.data.command = CMD_STOP;
            }

            if ( (last_cmd == CMD_PAUSE)&&
                 (sys.data.command != CMD_PAUSE) ) {
               broadcast_CONDVAR(sys.cv);
            }
            break;

         case DestroyNotify:
            process_events = 0;
         }
      }
   } while (process_events);
   exit_THREAD(NULL);
   return NULL;
}
#endif

#ifndef _NO_TERM_INPUT_
# define JUMP_OFFSET_INPUT_SIZE   16
static int parse_jump_offset(char *offset)
{
   double v1 = 0.0,
          v2 = 0.0,
          v3 = 0.0;
   int    rc = sscanf(offset, "%lf:%lf:%lf", &v1, &v2, &v3);

   if (rc == 3) {
      v1 *= 3600.0;
      v1 +=   60.0 * v2;
      v1 +=          v3;
   }
   else if (rc == 2) {
      v1 *=   60.0;
      v1 +=          v2;
   }
   
   if (v1 != 0.0)
      jump_request((CLK_VAL)(v1 * MPEG2_SYS_CLK_FREQ));
   else if (rc)
      sys.data.command = CMD_PLAY;
   return rc;
}

static void * THREAD_FUNC_TYPE TERMevent_loop(void *args)
{
   struct 
   termios org_state, new_state;
   u8      key[JUMP_OFFSET_INPUT_SIZE];
   u32     last_cmd,
           pos = 1,
          *run = (u32 *)args;
   
   if ( (isatty(STDIN_FILENO))&&(tcgetattr(STDIN_FILENO, &org_state) == 0) ) {
      memcpy(&new_state, &org_state, sizeof(struct termios));
      new_state.c_lflag    &= ~(ICANON | ECHO);
      new_state.c_cc[VMIN]  = 1;
      new_state.c_cc[VTIME] = 0;
      tcsetattr(STDIN_FILENO, TCSAFLUSH, &new_state);

#define KEY_SPACE 0x20
#define KEY_p     0x70
#define KEY_P     0x50
#define KEY_s     0x73
#define KEY_S     0x53
#define KEY_CTRLw 0x17
#define KEY_ESC   0x1B
#define KEY_EXT   0x5B
#define KEY_LEFT  0x44
#define KEY_UP    0x41
#define KEY_HOME  0x48
#define KEY_RIGHT 0x43
#define KEY_DOWN  0x42
#define KEY_PGUP  0x35
#define KEY_PGDWN 0x36
      key[1] = '\0';
      while (*run) {
         key[0] = 0;
         if (read(STDIN_FILENO, key, 1) > 0) {
            last_cmd = sys.data.command;

            switch (key[0]) {
            case KEY_SPACE:
            case KEY_p:
            case KEY_P:
               if (sys.data.command != CMD_PAUSE)
                  pause_request();
               else
                  sys.data.command = CMD_PLAY;
               break;
            case KEY_s:
            case KEY_S:
               if (sys.data.command != CMD_SLOW_FWD) {
                  if (sys.data.v_pos > 0)
                     sys.data.command = CMD_SLOW_FWD;
               }
               else
                  sys.data.command = CMD_PLAY;
               break;
            case KEY_ESC:
               new_state.c_cc[VMIN]  = 0;
               new_state.c_cc[VTIME] = 2;
               if (tcsetattr(STDIN_FILENO, (TCSANOW | TCSASOFT), &new_state)) {
                  printf("TERM ERROR: Cannot change attributes.\n");
                  break;
               }
               read(STDIN_FILENO, key, 1);
               new_state.c_cc[VMIN]  = 1;
               new_state.c_cc[VTIME] = 0;
               tcsetattr(STDIN_FILENO, (TCSANOW | TCSASOFT), &new_state);

               if (key[0] != KEY_ESC) {
                  if (key[0] == KEY_EXT)
                     read(STDIN_FILENO, key, 1);
                  switch (key[0]) {
                  case KEY_LEFT:
                     jump_request(-CMD_JUMP_SECS*MPEG2_SYS_CLK_FREQ);
                     break;
                  case KEY_UP:
                     jump_request(-CMD_JUMP2_SECS*MPEG2_SYS_CLK_FREQ);
                     break;
                  case KEY_HOME:
                     jump_request( 0);
                     break;
                  case KEY_RIGHT:
                     jump_request( CMD_JUMP_SECS*MPEG2_SYS_CLK_FREQ);
                     break;
                  case KEY_DOWN:
                     jump_request( CMD_JUMP2_SECS*MPEG2_SYS_CLK_FREQ);
#ifdef __DVD_INPUT__                     
                     break;
                  case KEY_PGUP:
                     jump_chapter_request(-1);
                     break;
                  case KEY_PGDWN:
                     jump_chapter_request( 1);
#endif
/*                     break;
                  default:
                     printf("0x1B%02hX\n", (u16)key[0]);*/
                  }
                  break;
               }
            case KEY_CTRLw:
               sys.data.command = CMD_STOP;
               break;
            default:
               if ( (key[0] == 0x08)&&(pos > 1) ) {
                  /* backspace */
                  key[--pos] = '\0';
                  printf("\r                      ");
                  goto TERMevent_loop_NEW_OFFSET_VALUE;
               }
               else if ( (key[0] == 0x0A)&&(parse_jump_offset(key+1)) ) {
                  /* return */
                  pos    = 1;
                  key[1] = '\0';
/*                  break;*/
               }
               else if (pos < (JUMP_OFFSET_INPUT_SIZE-1)) {
                  if ( (('0' <= key[0])&&(key[0] <= '9'))||
                        (key[0] == '+')||(key[0] == '-')||
                        (key[0] == ':')||(key[0] == '.') ) {
                     key[pos++] = key[0];
                     key[pos]   = '\0';
                     if (pos == 2) {
                        if (sys.data.command != CMD_PAUSE)
                           pause_request();
                        printf("\n");
                     }
TERMevent_loop_NEW_OFFSET_VALUE:
                     printf("\rJUMP> %s", (key+1));
                     fflush(stdout);
/*                     break;*/
                  }
               }
/*               printf("0x%02hX\n", (u16)key[0]);*/
            }

            if ( (last_cmd == CMD_PAUSE)&&
                 (sys.data.command != CMD_PAUSE) ) {
               broadcast_CONDVAR(sys.cv);
            }
         }
      }
#undef KEY_SPACE
#undef KEY_p
#undef KEY_P
#undef KEY_s
#undef KEY_S
#undef KEY_CTRLw
#undef KEY_ESC
#undef KEY_EXT
#undef KEY_LEFT
#undef KEY_UP
#undef KEY_HOME
#undef KEY_RIGHT
#undef KEY_DOWN
#undef KEY_PGUP
#undef KEY_PGDWN
      tcsetattr(STDIN_FILENO, TCSAFLUSH, &org_state);
   }
   exit_THREAD(NULL);
   return NULL;
}

static void usr1_sig(int sig)
{
   signal(SIGUSR1, SIG_DFL);
}

# undef JUMP_OFFSET_INPUT_SIZE
#endif

